
#include "stdafx.h"
#include "utils.h"
#include "stagetopics.h"
#include "loaders.h"
#include <omp.h>
///======================================================================
/// Main app.

int main(int argc, char* argv[]) {
	Env = TEnv(argc, argv, TNotify::StdNotify);
	Env.PrepArgs(TStr::Fmt("Inferring stages. build: %s, %s. Time: %s", __TIME__, __DATE__, TExeTm::GetCurTm()));
	TExeTm ExeTm;
	Try
	const TStr InFNm = Env.GetIfArgPrefixStr("-i:", "ba", "Input file name");
	const TStr INameFNm = Env.GetIfArgPrefixStr("-n:", "cname.txt", "Input file name");
  const int MinItems = Env.GetIfArgPrefixInt("-mi:", 50, "minimum number of items (length) for chain");
  const int MinChains = Env.GetIfArgPrefixInt("-mc:", 50, "minimum number of chains for items");
  const int NumThreads = Env.GetIfArgPrefixInt("-nt:", 4, "number of threads to use");
  const int NumTrials = Env.GetIfArgPrefixInt("-ni:", 5, "number of initializations to try");
  const int Stages = Env.GetIfArgPrefixInt("-st:", 5, "number of stages");
  const int Classes = Env.GetIfArgPrefixInt("-c:", 3, "classes of chains");
  const int RndSeed = Env.GetIfArgPrefixInt("-rs:", 0, "seed for random number generators");
  const int FitEM = Env.GetIfArgPrefixInt("-em:", 0, "Do EM or not (1:EM, 2:Greedy)");
  const double WDirichlet = Env.GetIfArgPrefixFlt("-wd:", 1, "Pseudo count for uniform Dirichlet prior");
	const TStr SettingFNm = Env.GetIfArgPrefixStr("-s:", "Settings.ini", "Setting file name");
  omp_set_num_threads(NumThreads);

  // Declare environmental variables
	TStrStrH SettingsH, GIDComsH;
	TYUtil::ReadStrStrH(SettingFNm, SettingsH, ssfTabSep);
  TStr DataPath;
	TStr OutPath = SettingsH.GetDat("OUTPATH"); 
  TStr OutFPrx;
  TStrHash<TInt> ChainIDH; // ID of chains
  TIntStrH IIDNameH; //Item Names
  THash<TInt, TFltIntPrV> CIDChainH; //Chain: <Time, Item ID> Vector
  TVec<TFltIntPrV> RawChainV; // Raw data
  TStr FitMtd = FitEM == 1? "em": "greedy";
  OutFPrx = OutPath + InFNm + TStr::Fmt(".%s.mi%d.mc%d.S%d.WD%.4f.C%d", FitMtd.CStr(), MinItems, MinChains, Stages, WDirichlet, Classes);
  //////////////////////////////////////////////////////////////////
  //Loading the data
  if (InFNm == "ba" || InFNm == "rb") { //Beer data
    DataPath = SettingsH.GetDat(TStr::Fmt("DATA_%s", InFNm.CStr())); 
    LoadBeer(DataPath, RawChainV, ChainIDH, IIDNameH);
  } else if (InFNm == "toy" || InFNm == "baby" || InFNm == "babyp" || InFNm == "movie" || InFNm == "book") {
    //Amazon data
    DataPath = SettingsH.GetDat(TStr::Fmt("DATA_%s", InFNm.CStr())); 
    // The following code saves temporary data into binary files for future experiments.
    TStr RCNm = DataPath.GetFPath() + InFNm + "_rawchain.save";
    TStrHash<TInt> ItemStrH;
    if (TFile::Exists(RCNm)) {
      TFIn CIn(RCNm);
      RawChainV.Load(CIn);
      TFIn IIn(TStr::PutFExt(RCNm, ".itemstr"));
      ItemStrH.Load(IIn);
    } else {
      LoadAmazon(DataPath, RawChainV, ChainIDH, ItemStrH);
      TFOut COut(RCNm);
      RawChainV.Save(COut);
      TFOut IOut(TStr::PutFExt(RCNm, ".itemstr"));
      ItemStrH.Save(IOut);
    }
    TStr ICFNm = DataPath.GetFPath() + "titles.txt";
    LoadAmazonTitle(ICFNm, IIDNameH, ItemStrH);
  }   else { //arbitrary input file
    if (InFNm.GetFExt() == ".save") {
    if (TFile::Exists(RCNm)) {
      TStr ICFNm = DataPath + "cname.txt";
      TYUtil::ReadIntStrH(ICFNm, IIDNameH, ssfTabSep);
      TFIn CIn(RCNm);
      RawChainV.Load(CIn);
    } else {
      DataFNm = DataPath + InFNm + "_notes.txt";
      LoadMed(DataFNm, RawChainV, ChainIDH, IIDNameH);
      TFOut COut(RCNm);
      RawChainV.Save(COut);
    }

    } else {
    //format: <Chain (User) ID> <Item 1> <Item 2> ... (no time input here)
    TSsParser Ss(InFNm, ssfWhiteSep);
    while (Ss.Next()) {
      if (Ss.GetFlds() < MinItems + 1) { continue; }
      TStr CStr = Ss.GetFld(0);
      ChainIDH.AddKey(CStr);
      TFltIntPrV ChainV;
      for (int i = 1; i < Ss.GetFlds(); i++) {
        int IID = Ss.GetInt(i);
        ChainV.Add(TFltIntPr((double) i, IID));
      }
      RawChainV.Add(ChainV);
    }
  }
  //Loading the data ends
  ///////////////////////////////////////////////////

  ApplyMinItemsChains(RawChainV, CIDChainH, MinChains, MinItems); //remove too infrequent items and too short chains
  /////////////////////////////////////////////////////
  //fitting
  TStageTopics ST(CIDChainH, Stages, Classes, RndSeed);
  ST.SetWDirichlet(WDirichlet);
  if (FitEM == 1) {
    ST.LearnEM(1000, 1e-4, "");
  } else {
    ST.FitBestL(NumTrials, 1000);
  }
  ST.DumpFullResults(OutFPrx + ".results.txt", ChainIDH);
  ST.DumpTopItems(OutFPrx + ".items.txt", 500, IIDNameH);
  Catch

  printf("\nrun time: %s (%s)\n", ExeTm.GetTmStr(), TSecTm::GetCurTm().GetTmStr().CStr());

  return 0;
}
