#ifndef yang_stagetopics_h
#define yang_stagetopics_h


#include "Snap.h"


struct KahanAccumulation
{
    double sum;
    double correction;
};

KahanAccumulation KahanSum(KahanAccumulation accumulation, double value)
{
    KahanAccumulation result;
    double y = value - accumulation.correction;
    double t = accumulation.sum + y;
    result.correction = (t - accumulation.sum) - y;
    result.sum = t;
    return result;
}

class TChain {
public:
  TIntV ItemsV;
  TIntV StagesV;
  TIntV ClassV;
  TFltV TimesV;
  TIntSet TestIdxS;
  //////// methods
  TChain(void) {}
  ~ TChain(void) {}
  TChain(const TFltIntPrV& ValIDV) {
    ItemsV.Gen(ValIDV.Len());
    StagesV.Gen(ValIDV.Len());
    ClassV.Gen(ValIDV.Len());
    TimesV.Gen(ValIDV.Len());
    TestIdxS.Gen(ValIDV.Len() / 5);
    IAssert(ValIDV.IsSorted(true));
    for (int i = 0; i < ValIDV.Len(); i++) {
      TimesV[i] = ValIDV[i].Val1;
      ItemsV[i] = ValIDV[i].Val2;
    }
  }
  TChain(const TChain& X): ItemsV(X.ItemsV), StagesV(X.StagesV), ClassV(X.ClassV), TimesV(X.TimesV), 
    TestIdxS(X.TestIdxS) {}
  TChain(TSIn& SIn): ItemsV(SIn), StagesV(SIn), ClassV(SIn), TimesV(SIn), TestIdxS(SIn) {}
  void Save(TSOut& SOut) const {
    ItemsV.Save(SOut);
    StagesV.Save(SOut);
    ClassV.Save(SOut);
    TimesV.Save(SOut);
    TestIdxS.Save(SOut);
  }
  void Load(TSIn& SIn) {
    ItemsV.Load(SIn);
    StagesV.Load(SIn);
    ClassV.Load(SIn);
    TimesV.Load(SIn);
    TestIdxS.Load(SIn);
  }
  int Len() const { return ItemsV.Len(); }
  void SetStage(const int IIdx, const int Stage) { IAssert(! IsTest(IIdx)); StagesV[IIdx] = Stage; }
  int GetStage(const int IIdx) const { IAssert(! IsTest(IIdx)); return StagesV[IIdx].Val; }
  double GetTime(const int IIdx) const { IAssert(! IsTest(IIdx)); return TimesV[IIdx].Val; }
  void SetClass(const int IIdx, const int Class) { IAssert(! IsTest(IIdx)); ClassV[IIdx] = Class; }
  int GetClass(const int IIdx) const { IAssert(! IsTest(IIdx)); return ClassV[IIdx].Val; }
  int GetClass() const { 
    for (int i = 0; i < ClassV.Len(); i++) { 
      if (! IsTest(i)) return GetClass(i); 
    } 
    return -1;
  }
  void SetClassAll(const int Class) {
    for (int i = 0; i < Len(); i++) { 
      if (IsTest(i)) { continue; }
      SetClass(i, Class); 
    }
  }
  void ClearTest() { TestIdxS.Clr(); }
  void PutTest(const int IIdx) { 
    //SetClass(IIdx, -1);
    SetStage(IIdx, -1);
    TestIdxS.AddKey(IIdx); 
  }
  bool IsTest(const int IIdx) const { return TestIdxS.IsKey(IIdx); }
  int GetClosestIdx(const int IIdx) {
    IAssert(IsTest(IIdx)); 
    int IdxBefore = - ItemsV.Len(), IdxAfter = 2 * ItemsV.Len();
    for (int j = IIdx - 1; j >= 0; j--) {
      if (! IsTest(j)) { IdxBefore = j; break; }
    }
    for (int j = IIdx + 1; j < ItemsV.Len(); j++) {
      if (! IsTest(j)) { IdxAfter = j; break; }
    }
    //printf("idx:%d, Len:%d, After:%d, Before:%d", IIdx, Len(), IdxAfter, IdxBefore);
    if (IdxAfter - IIdx <= IIdx - IdxBefore) {
      return IdxAfter;
    } else {
      return IdxBefore;
    }
  }
  int GetClassClosest(const int IIdx) {
    IAssert(IsTest(IIdx)); 
    return GetClass(GetClosestIdx(IIdx));
  }
  int GetStageClosest(const int IIdx) { 
    IAssert(IsTest(IIdx)); 
    return GetStage(GetClosestIdx(IIdx));
  }
  /*
  int LenForClass(const int Class) {
    int Len = 0;
    for (int i = 0; i < ItemsV.Len(); i++) {
      if (GetClass(i) == Class) { Len++; }
    }
    return Len;
  }
  */
  void ApplyPrior(const int IIdx, const int Class, const int Stage, const TIntSet& IIDS) {
    for (int i = 0; i < ItemsV.Len(); i++) {
      if (IsTest(i)) { continue; }
      if (GetClass(i) != Class) { continue; }
      if (GetItemIdx(i, IIDS) == IIdx && GetStage(i) != Stage) { 
        SetStage(i, Stage);
        for (int j = 0; j < i; j++) {
          if (IsTest(j)) { continue; }
          if (GetClass(j) == Class && GetStage(j) > Stage) { SetStage(j, Stage); }
        }
        for (int j = i + 1; j < ItemsV.Len(); j++) {
          if (IsTest(j)) { continue; }
          if (GetClass(j) == Class && GetStage(j) < Stage) { SetStage(j, Stage); }
        }
      }
    }
    IAssert(CheckStagesIncreasing());
  }
  void GetIdxForClass(TIntV& IdxV, const int Class) const{
    IdxV.Gen(ItemsV.Len(), 0);
    for (int i = 0; i < ItemsV.Len(); i++) {
      if (IsTest(i)) { continue; }
      if (GetClass(i) == Class) { IdxV.Add(i); }
    }
  }
  bool HasItemAtStage(const int IID, const int Stage, const int Class) {
    bool Result = false;
    for (int i = 0; i < ItemsV.Len(); i++) {
      if (IsTest(i)) { continue; }
      if (GetClass(i) != Class) { continue; }
      if (GetItemID(i) == IID && GetStage(i) == Stage) { 
        Result = true;
        break;
      }
    }
    return Result;
  }
  int GetItemID(const int IIdx) const { return ItemsV[IIdx].Val; }
  int GetItemIdx(const int IIdx, const TIntSet& IIDS) const { return IIDS.GetKeyId(GetItemID(IIdx)); }
  void Summary(const int TopI = -1) {
    int ItemsToShow = TopI > 0 && TopI < ItemsV.Len()? TopI: ItemsV.Len();
    printf("Class:%d\t", GetClass());
    for (int i = 0; i < ItemsToShow; i++) {
      if (IsTest(i)) {
        printf("%d(%d)", GetItemID(i), -1);
      } else {
        printf("%d(%d)", GetItemID(i), GetStage(i));
      }
      if (i < ItemsToShow - 1) { printf("-"); }
    }
    printf("\n");
  }
  bool CheckStagesIncreasing() { // check whether stages are always nondecreasing
    int MaxClasses = -1;
    for (int i = 0; i < ItemsV.Len(); i++) {
      if (IsTest(i)) { continue; }
      if (GetClass(i) + 1 > MaxClasses) { MaxClasses = GetClass(i) + 1; }
    }
    TIntV PrevStagesV(MaxClasses);
    for (int i = 0; i < ItemsV.Len(); i++) {
      if (IsTest(i)) { continue; }
      int Class = GetClass(i);
      if (PrevStagesV[Class] > GetStage(i)) { return false; }
      PrevStagesV[Class] = GetStage(i);
    }
    return true;
  }
};

class TChainUtil {
public:
  TChainUtil(void) {}
  ~ TChainUtil(void) {}
  static double SameCatLastItem(const TChain& Chain1, const TChain& Chain2, const THash<TInt, TIntSet>& IIDCatH) {
    if (Chain1.Len() == 0 || Chain2.Len() == 0) { return 0.0; }
    int IID1 = Chain1.GetItemID(Chain1.Len() - 1);
    int IID2 = Chain2.GetItemID(Chain2.Len() - 1);
    if (! IIDCatH.IsKey(IID1) || ! IIDCatH.IsKey(IID2)) { return 0.0; }
    int ComCat = TAGMUtil::Intersection(IIDCatH.GetDat(IID1), IIDCatH.GetDat(IID2));
    if (ComCat > 0) { return 1.0; }
    else { return 0.0; }
  }
};

class TStageTopics {
public:
  THash<TInt, TChain> CIDChainH;
  TInt Stages;
  TInt Classes;
  TInt Items;
  TIntSet IIDS;
  TFltVVV StageProbVVV;
  TFlt WDirichlet;
  TVec<TIntH> PriorIidxStageH;
  TInt RndSeed;
  //====================================================
  TStageTopics(const TStageTopics& X):CIDChainH(X.CIDChainH), Stages(X.Stages), Classes(X.Classes), Items(X.Items), 
    IIDS(X.IIDS), StageProbVVV(X.StageProbVVV), WDirichlet(X.WDirichlet), PriorIidxStageH(X.PriorIidxStageH), RndSeed(X.RndSeed) {}
  TStageTopics(void) {}
  ~ TStageTopics(void) {}
  TStageTopics(const THash<TInt, TFltIntPrV>& CIDTimeIDVH, const int _Stages, const int _Classes = 1, const int _Seed = 0) {
    CIDChainH.Gen(CIDTimeIDVH.Len());
    for (int c = 0; c < CIDTimeIDVH.Len(); c++) {
      TChain Chain(CIDTimeIDVH[c]);
      CIDChainH.AddDat(CIDTimeIDVH.GetKey(c), Chain);
      for (int i = 0; i < CIDTimeIDVH[c].Len(); i++) {
        IIDS.AddKey(CIDTimeIDVH[c][i].Val2);
      }
    }
    Items = IIDS.Len();
    Stages = _Stages;
    Classes = _Classes;
    RndSeed = _Seed;
    printf("%d chains loaded, %d items, %d stages\n", CIDChainH.Len(), Items.Val, Stages.Val);
    Items = IIDS.Len();
    WDirichlet = 1.0;
    StageProbVVV.Gen(Classes, Stages, Items);
    PriorIidxStageH.Gen(Classes);
    Init();
  }
  void Save(TSOut& SOut) {
    CIDChainH.Save(SOut);
    Stages.Save(SOut);
    Classes.Save(SOut);
    Items.Save(SOut);
    IIDS.Save(SOut);
    StageProbVVV.Save(SOut);
    WDirichlet.Save(SOut);
    PriorIidxStageH.Save(SOut);
    RndSeed.Save(SOut);
  }
  void Load(TSIn& SIn) {
    CIDChainH.Load(SIn);
    Stages.Load(SIn);
    Classes.Load(SIn);
    Items.Load(SIn);
    IIDS.Load(SIn);
    TFltVVV X(SIn);
    StageProbVVV = X;
    WDirichlet.Load(SIn);
    PriorIidxStageH.Load(SIn);
    RndSeed.Load(SIn);
  }
  double GetWDirichlet() { return WDirichlet; }
  void SetWDirichlet(const double w) { WDirichlet = w; }
  int GetItems() const { return Items; }
  void SetStages(const int _Stages) {
    Stages = _Stages;
    StageProbVVV.Gen(Classes, Stages, Items);
    Init();
  }
  double GetProbItem(const int Class, const int Stage, const int IIdx) {
    return StageProbVVV(Class, Stage, IIdx);
  }
  bool CheckProbs() {
    for (int l = 0; l < Classes; l++) {
      for (int s = 0; s < Stages; s++) {
        double SumProb = 0.0;
        for (int i = 0; i < Items; i++) {
          SumProb += GetProbItem(l, s, i);
        }
        if (fabs(SumProb - 1.0) > 1e-6) { 
          printf("Stage:%d P:%f\n", s, SumProb);
          return false; 
        }
      }
    }
    return true;
  }
  void SetProbItem(const int Class, const int Stage, const int IIdx, const double Val) {
    IAssert(Val <= 1.0);
    StageProbVVV(Class, Stage, IIdx) = Val;
  }
  void AddPriorStage(const int IID, const int Stage, const int Class = 0) {
    if (! IIDS.IsKey(IID)) { return; }
    IAssert(IIDS.IsKey(IID));
    int IIdx = IIDS.GetKeyId(IID);
    PriorIidxStageH[Class].AddDat(IIdx, Stage);
  }
  int GetTotalLen() {
    int TotalLen = 0;
    for (int c = 0; c < CIDChainH.Len(); c++) {
      TotalLen += CIDChainH[c].Len();
    }
    return TotalLen;
  }
  void Init(const int Seed = 0) {
    SetChainClass(Seed);
    InitializeStages();
    MLEProbs();
  }
  void InitNewClasses(const int _Classes, const int Seed = 0) {
    Classes = _Classes;
    Init(Seed);
  }
  void PutAllStageZero() {
    for (int c = 0; c < CIDChainH.Len(); c++) {
      TChain& Chain = CIDChainH[c];
      for (int i = 0; i < Chain.Len(); i++) {
        if (Chain.IsTest(i)) { continue; }
        Chain.SetStage(i, 0);
      }
    }
    MLEProbs();
  }
  void ReserveZeroStage() {
    Stages = Stages + 1;
    StageProbVVV.Gen(Classes, Stages, Items);

    for (int c = 0; c < CIDChainH.Len(); c++) {
      TChain& Chain = CIDChainH[c];
      if (Chain.Len() == 0) { continue; }
      if (! Chain.IsTest(0)) { Chain.SetStage(0, 0); }
      for (int i = 1; i < Chain.Len(); i++) {
        if (Chain.IsTest(i)) { continue; }
        Chain.SetStage(i, Chain.GetStage(i) + 1);
      }
    }
    MLEProbs();
  }
  void SetChainClass(const int Seed = 0) {
    TRnd Rnd(Seed);
    for (int c = 0; c < CIDChainH.Len(); c++) {
      TChain& Chain = CIDChainH[c];
      int Class = Rnd.GetUniDevInt(Classes);
      //int Class = c % Classes;
      Chain.SetClassAll(Class);
    }
  }
  void InitializeStages() {
    IAssert(Stages > 0);
    for (int c = 0; c < CIDChainH.Len(); c++) {
      TChain& Chain = CIDChainH[c];
      for (int l = 0; l < Classes; l++) {
        TIntV IIdxV;
        Chain.GetIdxForClass(IIdxV, l);
        if (IIdxV.Len() == 0) { continue; }
        int StageLen = (int) TMath::Round((double) IIdxV.Len() / (double) Stages);
        for (int i = 0; i < IIdxV.Len(); i++) {
          int SIdx = (int) floor((double) i / StageLen);
          if (SIdx >= Stages) { SIdx = Stages - 1; }
          Chain.SetStage(IIdxV[i], SIdx);
        }
        for (int l = 0; l < Classes; l++) {
          for (int p = 0; p < PriorIidxStageH[l].Len(); p++) {
            int IIdx = PriorIidxStageH[l].GetKey(p);
            int Stage = PriorIidxStageH[l][p];
            Chain.ApplyPrior(IIdx, l, Stage, IIDS);
          }
        }
      }
      IAssert(Chain.CheckStagesIncreasing());
    }
    printf("stages initialized\n");
  }
  void MLEProbs() {
    for (int l = 0; l < Classes; l++) {
#pragma omp parallel for
      for (int s = 0; s < Stages; s++) {
        TIntV ItemCntV(Items);
        int TotalCnt = 0;
        for (int c = 0; c < CIDChainH.Len(); c++) {
          TChain& Chain = CIDChainH[c];
          for (int i = 0; i < Chain.Len(); i++) {
            if (Chain.IsTest(i)) { continue; }
            IAssert(Chain.GetStage(i) < Stages);
            if (Chain.GetClass(i) != l) { continue; }
            //if (Chain.GetClass(i) != l) { break; }
            if (Chain.GetStage(i) > s) { break; }
            if (Chain.GetStage(i) < s) { continue; }
            int Item = Chain.GetItemIdx(i, IIDS);
            IAssert(Item < Items);
            /*
            if (PriorIidxStageH[l].IsKey(Item)) {
              IAssert(PriorIidxStageH[l].GetDat(Item) == s); 
            }
            */
            ItemCntV[Item]++;
            TotalCnt++;
          }
        }
        int ItemWithSupport = Items;
        for (int i = 0; i < PriorIidxStageH[l].Len(); i++) { // if some items belong to other stages by prior, remove their contribution for the dirichlet prior
          if (PriorIidxStageH[l][i] == s) { continue; }
          //ItemWithSupport--;
        }
        //printf("Stage:%d, Items:%d, With support:%d\n", s, Items, ItemWithSupport);
        for (int i = 0; i < Items; i++) {
          if (PriorIidxStageH[l].IsKey(i) && PriorIidxStageH[l].GetDat(i) != s) {
            SetProbItem(l, s, i, 0.00001 * WDirichlet / (double) ItemWithSupport);
          } else {
            SetProbItem(l, s, i, ((double) ItemCntV[i] + WDirichlet) / ((double) TotalCnt + WDirichlet * ItemWithSupport));
          }
        }
      }
    }
    //IAssert(CheckProbs());
  }
  void GetAvgProb(TFltV& ProbV, const int Class) {
    ProbV.Gen(Items);
    TIntV ItemCntV(Items);
    int TotalCnt = 0;
    for (int c = 0; c < CIDChainH.Len(); c++) {
      TChain& Chain = CIDChainH[c];
      for (int i = 0; i < Chain.Len(); i++) {
        if (Chain.IsTest(i)) { continue; }
        if (Chain.GetClass(i) != Class) { continue; }
        int Item = Chain.GetItemIdx(i, IIDS);
        IAssert(Item < Items);
        ItemCntV[Item]++;
        TotalCnt++;
      }
    }
    //printf("Stage:%d, Items:%d, With support:%d\n", s, Items, ItemWithSupport);
    for (int i = 0; i < Items; i++) {
      ProbV[i] = ((double) ItemCntV[i] + WDirichlet) / ((double) TotalCnt + WDirichlet * Items);
    }
  }
  void DumpFullResults(const TStr FNm, const TStrHash<TInt>& CNameH) {
    FILE* F = fopen(FNm.CStr(), "wt");
    fprintf(F, "#<Chain(User) ID>\t<Class>\t<Item1>-<Stage1>\t<Item2>-<Stage2>-...\n");
    for (int c = 0; c < CIDChainH.Len(); c++) {
      TChain& Chain = CIDChainH[c];
      int Class = Chain.GetClass();
      int CID = CIDChainH.GetKey(c);
      TStr CStr = CNameH.Len() == 0? TStr::Fmt("%d", CID) : CNameH.GetKey(CID);
      fprintf(F, "%s\t%d", CStr.CStr(), Class);
      for (int i = 0; i < Chain.Len(); i++) {
        fprintf(F, "\t%d-%d", Chain.GetItemID(i), Chain.GetStage(i));
      }
      fprintf(F, "\n");
    }
    fclose(F);
  }
  void GetLogisticTrainingSet(TVec<TIntFltH>& XV, TIntV& YV, const int MultForTest, const int MaxRows = -1) {
    //split the training set into x and y
    for (int c = 0; c < CIDChainH.Len(); c++) {
      TChain& Chain = CIDChainH[c];
      TIntSet YIdx;
      for (int i = 0; i < Chain.Len(); i++) {
        if (Chain.IsTest(i)) {
          int YAdded = 0;
          for (int j = i - 1; j > 0; j--) {
            if (! YIdx.IsKey(j) && ! Chain.IsTest(j)) {
              YIdx.AddKey(j);
              YAdded++;
            }
            if (YAdded == MultForTest) { break; }
          }
        }
      }
      // add x
      TIntFltH X;
      for (int i = 0; i < Chain.Len(); i++) {
        if (! YIdx.IsKey(i) && ! Chain.IsTest(i)) {
          int IIdx = Chain.GetItemIdx(i, IIDS);
          if (! X.IsKey(IIdx)) { X.AddDat(IIdx, 0.0); }
          //X.GetDat(IIdx) += 1.0;
          X.GetDat(IIdx) = 1.0;
        }
      }
      X.SortByKey();
      for (int y = 0; y < YIdx.Len(); y++) {
        int YVal = Chain.GetItemIdx(YIdx[y], IIDS);
        YV.Add(YVal);
        XV.Add(X);
      }
    }
    if (MaxRows > 0 && MaxRows < XV.Len()) {
      TIntV IdxV;
      TRnd Rnd(MaxRows);
      for (int i = 0; i < XV.Len(); i++) { IdxV.Add(i); }
      IdxV.Shuffle(Rnd);
      TVec<TIntFltH> NewXV(XV.Len(), 0);
      TIntV NewYV(YV.Len(), 0);
      for (int i = 0; i < MaxRows; i++) {
        int Idx = IdxV[i];
        NewXV.Add(XV[Idx]);
        NewYV.Add(YV[Idx]);
      }
      XV = NewXV;
      YV = NewYV;
    }
  }
  void GetLogisticTestSet(TVec<TIntFltH>& XV, TIntV& YV) {
    //split the training set into x and y
    for (int c = 0; c < CIDChainH.Len(); c++) {
      TChain& Chain = CIDChainH[c];
      TIntSet YIdx;
      for (int i = 0; i < Chain.Len(); i++) {
        if (Chain.IsTest(i)) {
          YIdx.AddKey(i);
        }
      }
      // add x
      TIntFltH X;
      for (int i = 0; i < Chain.Len(); i++) {
        if (! YIdx.IsKey(i) && ! Chain.IsTest(i)) {
          int IIdx = Chain.GetItemIdx(i, IIDS);
          if (! X.IsKey(IIdx)) { X.AddDat(IIdx, 0.0); }
          //X.GetDat(IIdx) += 1.0;
          X.GetDat(IIdx) = 1.0;
        }
      }
      X.SortByKey();
      for (int y = 0; y < YIdx.Len(); y++) {
        int YVal = Chain.GetItemIdx(YIdx[y], IIDS);
        YV.Add(YVal);
        XV.Add(X);
      }
    }
  }
  void AccLogistic(double& Acc, double& L, const int YFactor, const int TopK, const TStr FPrx, const TStr LibPath, int& ProbSize, const int MaxRows = -1) {
    printf("logistic regression begins\n");
    time_t InitTime = time(NULL);
    time_t LapTime = time(NULL);
    TVec<TIntFltH> XTrain, XTest, ProbHV;
    TIntV YTrain, YTest, YEst;
    GetLogisticTrainingSet(XTrain, YTrain, YFactor, MaxRows);
    GetLogisticTestSet(XTest, YTest);
    ProbSize = XTrain.Len();
    printf("training set (%d examples), test set generated [%d secs]\n", XTrain.Len(), time(NULL) - InitTime);
    LapTime = time(NULL);
    TStr TrainFNm = FPrx + ".train";
    TStr TestFNm = FPrx + ".test";
    TStr ModelFNm = FPrx + ".model";
    TStr OutFNm = FPrx + ".out";
    TYUtil::LibLinearTrain(TrainFNm, ModelFNm, LibPath + "train", XTrain, YTrain, "-s 0");
    printf("training done [%d secs]\n", time(NULL) - LapTime);
    LapTime = time(NULL);
    TYUtil::LibLinearTest(TestFNm, ModelFNm, OutFNm, LibPath + "predict", XTest, YTest, ProbHV, "-b 1");
    printf("test done [%d secs]\n", time(NULL) - LapTime);
    L = TYUtil::MeasureMultLikelihood(YTest, ProbHV, WDirichlet);
    Acc = TYUtil::MeasureAccTopK(YTest, ProbHV, TopK);
    system(TStr::Fmt("rm %s", TrainFNm.CStr()).CStr());
    system(TStr::Fmt("rm %s", TestFNm.CStr()).CStr());
    system(TStr::Fmt("rm %s", ModelFNm.CStr()).CStr());
    system(TStr::Fmt("rm %s", OutFNm.CStr()).CStr());
  }
  void Summary(int TopChains = -1, int TopItems = -1) {
    int ChainsToShow = TopChains > 0 && TopChains < CIDChainH.Len()? TopChains: CIDChainH.Len(); 
    for (int c = 0; c < ChainsToShow; c++) {
      CIDChainH[c].Summary(TopItems);
    }
  }
  void SummaryProbs(int TopItems = -1) {
    int ItemsToShow = TopItems > 0 && TopItems < Items? TopItems: Items.Val; 
    for (int l = 0; l < Classes; l++) {
      printf("Class %d\n", l);
      for (int s = 0; s < Stages; s++) {
        TIntFltH IIdxProbH;
        for (int i = 0; i < Items; i++) {
          IIdxProbH.AddDat(i, GetProbItem(l, s, i));
        }
        IIdxProbH.SortByDat(false);
        printf("Stage %d: ", s);
        for (int i = 0; i < ItemsToShow; i++) {
          int IIdx = IIdxProbH.GetKey(i);
          printf("%d(%.2f)\t", IIDS.GetKey(IIdx).Val, IIdxProbH[i].Val);
        }
        printf("\n");
      }
    }
  }
  void GetAvgLenPerStage(TFltV& LenV, const int Class) {
    LenV.Gen(Stages);
    LenV.PutAll(0.0);
    TIntV CurLenV(Stages);
    int ChainCnt = 0;
    for (int c = 0; c < CIDChainH.Len(); c++) {
      TChain& Chain = CIDChainH[c];
      TIntV IdxV;
      CurLenV.PutAll(0);
      Chain.GetIdxForClass(IdxV, Class);
      if (IdxV.Len() == 0) { continue; }
      for (int i = 0; i < IdxV.Len(); i++) {
        CurLenV[Chain.GetStage(IdxV[i])]++;
      }
      for (int s = 0; s < Stages; s++) {
        LenV[s] += ((double) CurLenV[s] / (double) IdxV.Len());
      }
      ChainCnt++;
    }
    for (int s = 0; s < Stages; s++) {
      LenV[s] /= (double) ChainCnt;
    }
  }
  void GetItemsPerStage(TFltV& LenV, const int Class) {
    LenV.Gen(Stages);
    LenV.PutAll(0.0);
    int TotalLen = 0;
    for (int c = 0; c < CIDChainH.Len(); c++) {
      TChain& Chain = CIDChainH[c];
      TIntV IdxV;
      Chain.GetIdxForClass(IdxV, Class);
      if (IdxV.Len() == 0) { continue; }
      for (int i = 0; i < IdxV.Len(); i++) {
        LenV[Chain.GetStage(IdxV[i])]++;
      }
      TotalLen += IdxV.Len();
    }
    for (int s = 0; s < Stages; s++) {
      LenV[s] /= (double) TotalLen;
    }
  }
  void GetTopKItems(const int TopK, TVVec<TIntSet>& StageTopItemH) {
    StageTopItemH.Gen(Classes, Stages);
    for (int l = 0; l < Classes; l++) {
      for (int s = 0; s < Stages; s++) {
        TIntFltH IIDProbH;
        for (int i = 0; i < Items; i++) {
          IIDProbH.AddDat(IIDS.GetKey(i), GetProbItem(l, s, i));
        }
        IIDProbH.SortByDat(false);
        for (int i = 0; i < TopK; i++) {
          StageTopItemH(l, s).AddKey(IIDProbH.GetKey(i));
        }
      }
    }
  }
  double GetAccPredictingTopK(const int TopK) {
    TVVec<TIntSet> StageTopItemH;
    GetTopKItems(TopK, StageTopItemH);
    double SumAcc = 0.0;
    int Cnt = 0, Hits = 0;
    for (int c = 0; c < CIDChainH.Len(); c++) {
      TChain& Chain = CIDChainH[c];
      //int Hits = 0, TestLen = 0;
      for (int i = 0; i < Chain.Len(); i++) {
        if (! Chain.IsTest(i)) { continue; }
        int IID = Chain.GetItemID(i);
        int Stage = Chain.GetStageClosest(i);
        int Class = Chain.GetClass();
        //TestLen++;
        if (StageTopItemH(Class, Stage).IsKey(IID)) { Hits++; }
        Cnt++;
      }
      //SumAcc += (double) Hits / (double) TestLen;
      //Cnt++;
    }
    return (double) Hits / (double) Cnt;
  }

  double GetF1PredictingTopK(const int TopK) {
    TVVec<TIntSet> StageTopItemH;
    GetTopKItems(TopK, StageTopItemH);
    double SumF = 0.0;
    int Cnt = 0;
    for (int c = 0; c < CIDChainH.Len(); c++) {
      TChain& Chain = CIDChainH[c];
      TIntSet GTIIDS;
      IAssert(Chain.IsTest(Chain.Len() - 1));
      for (int i = 0; i < Chain.Len(); i++) {
        if (Chain.IsTest(i)) { GTIIDS.AddKey(Chain.GetItemID(i)); }
      }
      if (GTIIDS.Len() == 0) { continue; }
      int Stage = Chain.GetStageClosest(Chain.Len() - 1);
      int Class = Chain.GetClass();
      SumF += TYUtil::F1Score(GTIIDS, StageTopItemH(Class, Stage));
      Cnt++;
    }
    return SumF / (double) Cnt;
  }
  void GetLenStatsPerClass(TVec<TMom>& MomV) {
    MomV.Gen(Classes);
    for (int c = 0; c < CIDChainH.Len(); c++) {
      int Class = CIDChainH[c].GetClass();
      MomV[Class].Add(CIDChainH[c].Len());
    }
    for (int l = 0; l < Classes; l++) {
      MomV[l].Def();
    }
  }
  void DumpTopItems(const TStr FNm, int TopItems = -1, const TIntStrH& INameH = TIntStrH()) {
    int ItemsToShow = TopItems > 0 && TopItems < Items? TopItems: Items.Val; 
    TIntV CntV;
    GetChainClasses(CntV);
    TVec<TMom> LenMomV;
    GetLenStatsPerClass(LenMomV);
    for (int l = 0; l < Classes; l++) {
      if (CntV[l] == 0) { continue; }
      TVec<TIntFltH> StageTopItemH(Stages);
      for (int s = 0; s < Stages; s++) {
        TIntFltH IIdxProbH;
        for (int i = 0; i < Items; i++) {
          IIdxProbH.AddDat(i, GetProbItem(l, s, i));
        }
        IIdxProbH.SortByDat(false);
        StageTopItemH[s] = IIdxProbH;
      }
      TFltV LenV, ItemsV;
      GetAvgLenPerStage(LenV, l);
      GetItemsPerStage(ItemsV, l);
      TMom& LenMom = LenMomV[l];
      FILE* F = fopen((FNm + TStr::Fmt(".c%d.topitems.txt", l)).CStr(), "wt");
      fprintf(F, "class %d, %d chains\n", l, CntV[l].Val);
      fprintf(F, "AvgLen:%.1f\tMedian:%.1f\n\n", LenMom.GetMean(), LenMom.GetMedian());
      for (int s = 0; s < Stages; s++) {
        fprintf(F, "Avg Len\t%.2f", LenV[s].Val);
        if (s < Stages - 1) { fprintf(F, "\t"); }
      }
      fprintf(F, "\n");
      for (int s = 0; s < Stages; s++) {
        fprintf(F, "Items\t%.2f", ItemsV[s].Val);
        if (s < Stages - 1) { fprintf(F, "\t"); }
      }
      fprintf(F, "\n");
      for (int i = 0; i < ItemsToShow; i++) {
        for (int s = 0; s < Stages; s++) {
          int IID = IIDS.GetKey(StageTopItemH[s].GetKey(i)).Val;
          TStr IStr = INameH.IsKey(IID)? INameH.GetDat(IID): TStr::Fmt("%d", IID);
          fprintf(F, "%s\t%f", IStr.CStr(), StageTopItemH[s][i].Val);
          if (s < Stages - 1) {
            fprintf(F, "\t");
          }
        }
        fprintf(F, "\n");
      }
      fclose(F);
    }
  }
  void DumpClusters(const TStr FNm, const TStrHash<TInt>& CNameH) {
    TVec<TIntH> CIDForClass(Classes);
    TIntV CntV(Classes);
    for (int c = 0; c < CIDChainH.Len(); c++) {
      int Class = CIDChainH[c].GetClass();
      CIDForClass[Class].AddDat(CIDChainH.GetKey(c), CIDChainH[c].Len());
      CntV[Class]++;
    }
    for (int l = 0; l < Classes; l++) {
      TIntH& CIDLenH = CIDForClass[l];
      CIDLenH.SortByDat(false);
      FILE* F = fopen((FNm + TStr::Fmt(".c%d.members.txt", l)).CStr(), "wt");
      fprintf(F, "class %d, %d chains\n\n", l, CntV[l].Val);
      for (int c = 0; c < CIDLenH.Len(); c++) {
        int CID = CIDLenH.GetKey(c);
        int Len = CIDLenH[c];
        IAssert(CNameH.Len() > CID);
        TStr IStr = CNameH.GetKey(CID);
        fprintf(F, "%s\t%d\n", IStr.CStr(), Len);
      }
      fclose(F);
    }
  }
  void DumpTopRelItems(const TStr FNm, int TopItems = -1, const TIntStrH& INameH = TIntStrH()) {
    int ItemsToShow = TopItems > 0 && TopItems < Items? TopItems: Items.Val; 
    TIntV CntV;
    GetChainClasses(CntV);
    for (int l = 0; l < Classes; l++) {
      if (CntV[l] == 0) { continue; }
      TFltV ProbV;
      TVec<TIntFltH> StageTopItemH(Stages);
      GetAvgProb(ProbV, l);
      for (int s = 0; s < Stages; s++) {
        TIntFltH IIdxProbH;
        for (int i = 0; i < Items; i++) {
          IIdxProbH.AddDat(i, GetProbItem(l, s, i) - ProbV[i]);
        }
        IIdxProbH.SortByDat(false);
        StageTopItemH[s] = IIdxProbH;
      }
      TFltV LenV, ItemsV;
      GetAvgLenPerStage(LenV, l);
      GetItemsPerStage(ItemsV, l);
      FILE* F = fopen((FNm + TStr::Fmt(".c%d.relitems.txt", l)).CStr(), "wt");
      fprintf(F, "class %d, %d chains\n\n", l, CntV[l].Val);
      for (int s = 0; s < Stages; s++) {
        fprintf(F, "Avg Len\t%.2f", LenV[s].Val);
        if (s < Stages - 1) { fprintf(F, "\t"); }
      }
      fprintf(F, "\n");
      for (int s = 0; s < Stages; s++) {
        fprintf(F, "Items\t%.2f", ItemsV[s].Val);
        if (s < Stages - 1) { fprintf(F, "\t"); }
      }
      fprintf(F, "\n");
      for (int i = 0; i < ItemsToShow; i++) {
        for (int s = 0; s < Stages; s++) {
          int IID = IIDS.GetKey(StageTopItemH[s].GetKey(i)).Val;
          TStr IStr = INameH.IsKey(IID)? INameH.GetDat(IID): TStr::Fmt("%d", IID);
          fprintf(F, "%s\t%.3f", IStr.CStr(), StageTopItemH[s][i].Val);
          if (s < Stages - 1) {
            fprintf(F, "\t");
          }
        }
        fprintf(F, "\n");
      }
      fclose(F);
    }
  }
  /*
  void DumpTopRelItemsAvgChain(const TStr FNm, int TopItems = -1, const TIntStrH& INameH = TIntStrH()) {
    int ItemsToShow = TopItems > 0 && TopItems < Items? TopItems: Items; 
    TFltV ProbV;
    TVVec<TIntFltH> StageTopItemH(Classes, Stages);
    for (int l = 0; l < Classes; l++) {
      GetAvgProb(ProbV, l);
      for (int s = 0; s < Stages; s++) {
        TIntFltH IIdxProbH;
        for (int i = 0; i < Items; i++) {
          //IIdxProbH.AddDat(i, GetProbItem(l, s, i) - SumProbV(l, i) / (double) Stages);
          IIdxProbH.AddDat(i, GetProbItem(l, s, i) - ProbV[i]);
        }
        IIdxProbH.SortByDat(false);
        StageTopItemH(l, s) = IIdxProbH;
      }
    }
    FILE* F = fopen(FNm.CStr(), "wt");
    for (int i = 0; i < ItemsToShow; i++) {
      for (int l = 0; l < Classes; l++) {
        for (int s = 0; s < Stages; s++) {
          int IID = IIDS.GetKey(StageTopItemH(l, s).GetKey(i)).Val;
          TStr IStr = INameH.IsKey(IID)? INameH.GetDat(IID): TStr::Fmt("%d", IID);
          fprintf(F, "%s\t%.3f", IStr.CStr(), StageTopItemH(l, s)[i].Val);
          if (s < Stages - 1 || l < Classes - 1) {
            fprintf(F, "\t");
          }
        }
      }
      fprintf(F, "\n");
    }
    fclose(F);
  }
  */
  int GetMLEStageByProb(const int IIdx, double& MaxP, double& SumProbs, const int Class) {
    TFltV ProbV(Stages);
    for (int s = 0; s < Stages; s++) {
      ProbV[s] = GetProbItem(Class, s, IIdx);//StageProbV[s][IIdx];
    }
    int MaxStage = -1; 
    MaxP = 0.0;
    SumProbs = TLinAlg::SumVec(ProbV);
    for (int s = 0; s < ProbV.Len(); s++) {
      if (ProbV[s] > MaxP) { 
        MaxStage = s;
        MaxP = ProbV[s];
      }
    }
    return MaxStage;
  }
  int GetMLEStageByProb(const int IIdx, const int Class = 0) {
    double Tmp1, Tmp2;
    return GetMLEStageByProb(IIdx, Tmp1, Tmp2, Class);
  }
  int GetMLEStageProbForIID(const int IID, const int Class = 0) {
    IAssert(IIDS.IsKey(IID));
    int IIdx = IIDS.GetKeyId(IID);
    return GetMLEStageByProb(IIdx, Class);
  }
  void GetMLEStageInferred(TIntFltH& IIDStageH) {
    IIDStageH.Gen(Items);
    TIntH IIDCntH(Items);
    for (int c = 0; c < CIDChainH.Len(); c++) {
      TChain& Chain = CIDChainH[c];
      for (int i = 0; i < Chain.Len(); i++) {
        if (Chain.IsTest(i)) { continue; }
        int IID = Chain.GetItemID(i);
        //if (Chain.GetClass(i) != Class) { continue; }
        int Stage = Chain.GetStage(i);
        IIDStageH.AddDat(IID) += (double) Stage;
        IIDCntH.AddDat(IID)++;
      }
    }
    for (int i = 0; i < IIDCntH.Len(); i++) {
      int IID = IIDCntH.GetKey(i);
      IIDStageH.GetDat(IID) /= (double) IIDCntH.GetDat(IID);
    }
  }
  double AccCausalWithProb(const TIntH& CPairH, int& NumMatches, const int Class) {
    double AccCnt = 0.0;
    NumMatches = 0;
    for (int i = 0; i < CPairH.Len(); i++) {
      int CauseID = CPairH.GetKey(i);
      int ResultID = CPairH[i];
      if (! IIDS.IsKey(CauseID) || ! IIDS.IsKey(ResultID)) { continue; }
      int CauseStage = GetMLEStageProbForIID(CauseID, Class);
      int ResultStage = GetMLEStageProbForIID(ResultID, Class);
      if (CauseStage < ResultStage) { AccCnt += 1.0; }
      //if (CauseStage == ResultStage) { AccCnt += 0.5; }
      NumMatches++;
    }
    return AccCnt / (double) NumMatches;
  }
  double AccCausalWithStage(const TIntH& CPairH, int& NumMatches) {
    double AccCnt = 0.0;
    NumMatches = 0;
    TIntFltH IIDStageH;
    GetMLEStageInferred(IIDStageH);
    for (int i = 0; i < CPairH.Len(); i++) {
      int CauseID = CPairH.GetKey(i);
      int ResultID = CPairH[i];
      if (! IIDStageH.IsKey(CauseID) || ! IIDStageH.IsKey(ResultID)) { continue; }
      double CauseStage = IIDStageH.GetDat(CauseID);
      double ResultStage = IIDStageH.GetDat(ResultID);
      if (CauseStage < ResultStage) { AccCnt += 1.0; }
      //if (CauseStage == ResultStage) { AccCnt += 0.5; }
      NumMatches++;
    }
    return AccCnt / (double) NumMatches;
  }
  double CorrGTWithProb(const TIntFltH& IIDGTH, int& NumMatches, const int Class, const TIntStrH& IIDNameH, const bool RankCorr = false, const TStr FNm = "") {
    TFltPrV EstGTV;
    FILE* F;
    if (FNm != "") {
      F = fopen(FNm.CStr(), "wt");
    }
    for (int i = 0; i < IIDGTH.Len(); i++) {
      int IID = IIDGTH.GetKey(i);
      if (! IIDS.IsKey(IID)) { continue; }
      //printf("ID found (id: %d, idx:%d)\n", IID, IIDS.GetKeyId(IID));
      double MLEStage = (double) GetMLEStageProbForIID(IID, Class);
      //printf("MLE stage found (%f, %f)\n", IIDGTH[i].Val, MLEStage);
      EstGTV.Add(TFltPr(IIDGTH[i], MLEStage));
      TStr IName = IIDNameH.IsKey(IID)? IIDNameH.GetDat(IID): TStr::Fmt("%d", IID);
      if (FNm != "") {
        fprintf(F, "%d\t%s\t%.2f\t%.2f\n", IID, IName.CStr(), IIDGTH[i].Val, MLEStage);
      }
    }
    if (FNm != "") { fclose(F);  }
    printf("done\n\n");
    
    NumMatches = EstGTV.Len();
    printf("Correlation computed from %d pairs\n", EstGTV.Len());
    double Corr;
    if (RankCorr) {
      Corr = TYUtil::RankCorr(EstGTV);
    } else {
      Corr = TYUtil::Corr(EstGTV);
    }
    return Corr;
  }
  double CorrGTWithStage(const TIntFltH& IIDGTH, int& NumMatches, const TIntStrH& IIDNameH, const bool RankCorr = false, const TStr FNm = "") {
    TFltPrV EstGTV;
    FILE* F;
    if (FNm != "") { F = fopen(FNm.CStr(), "wt"); }
    TIntFltH IIDStageH;
    GetMLEStageInferred(IIDStageH);
    for (int i = 0; i < IIDGTH.Len(); i++) {
      int IID = IIDGTH.GetKey(i);
      if (! IIDStageH.IsKey(IID)) { continue; }
      double MLEStage = (double) IIDStageH.GetDat(IID);
      //printf("MLE stage found (%f, %f)\n", IIDGTH[i].Val, MLEStage);
      EstGTV.Add(TFltPr(IIDGTH[i], MLEStage));
      TStr IName = IIDNameH.IsKey(IID)? IIDNameH.GetDat(IID): TStr::Fmt("%d", IID);
      if (FNm != "") {
        fprintf(F, "%d\t%s\t%.2f\t%.2f\n", IID, IName.CStr(), IIDGTH[i].Val, MLEStage);
      }
    }
    if (FNm != "") { fclose(F);  }
    printf("done\n\n");
    NumMatches = EstGTV.Len();
    printf("Correlation computed from %d pairs\n", EstGTV.Len());
    double Corr;
    if (RankCorr) {
      Corr = TYUtil::RankCorr(EstGTV);
    } else {
      Corr = TYUtil::Corr(EstGTV);
    }
    return Corr;
  }
  void GetIIDCnt(TIntH& IIDCntH) {
    IIDCntH.Gen(Items);
    for (int c = 0; c < CIDChainH.Len(); c++) {
      TChain& Chain = CIDChainH[c];
      for (int i = 0; i < Chain.Len(); i++) {
        IIDCntH.AddDat(Chain.GetItemID(i))++;
      }
    }
  }
  void GetIIDCntForClass(TIntH& IIDCntH, const int Class) {
    IIDCntH.Gen(Items);
    for (int c = 0; c < CIDChainH.Len(); c++) {
      TChain& Chain = CIDChainH[c];
      if (Chain.GetClass() != Class) { continue; }
      for (int i = 0; i < Chain.Len(); i++) {
        IIDCntH.AddDat(Chain.GetItemID(i))++;
      }
    }
  }

  void DumpItemStage(const TStr FNm, const TIntStrH& INameH = TIntStrH()) {
    TIntH IIDCntH;
    GetIIDCnt(IIDCntH);
    FILE* F = fopen(FNm.CStr(), "wt");
    for (int i = 0; i < Items; i++) {
      int IID = IIDS.GetKey(i);
      TStr IStr = INameH.IsKey(IID)? INameH.GetDat(IID): TStr::Fmt("%d", IID);
      fprintf(F, "%d\t%s\t%d", IID, IStr.CStr(), IIDCntH.GetDat(IID).Val);
      for (int Class = 0; Class < Classes; Class++) {
        TIntH IIDCntClassH;
        GetIIDCntForClass(IIDCntClassH, Class);
        double MaxP, SumProbs;
        int MaxStage = GetMLEStageByProb(i, MaxP, SumProbs, Class);
        double MaxRelP = SumProbs >= 0.0? MaxP / SumProbs: 0.0;
        int ICnt = IIDCntClassH.IsKey(IID)? IIDCntClassH.GetDat(IID).Val : 0;
        fprintf(F, "\t%d\t%d\t%.2f", ICnt, MaxStage, MaxRelP);
        TFltV ProbV(Stages);
        for (int s = 0; s < Stages; s++) {
          ProbV[s] = GetProbItem(Class, s, i);//StageProbV[s][i];
        }
        for (int s = 0; s < Stages; s++) {
          fprintf(F, "\t%f", ProbV[s].Val);
        }
      }
      fprintf(F, "\n");
    }
    fclose(F);
  }
  double GetScoreForItem(const TChain& Chain, const int IIdx, const int Stage, const int Class) {
    IAssertR(Stage >= 0 && Stage < Stages, TStr::Fmt("stage:%d", Stage));
    IAssertR(Class >= 0 && Class < Classes, TStr::Fmt("Class:%d", Class));
    IAssertR(IIdx >= 0 && IIdx < Chain.Len(), TStr::Fmt("IIdx:%d", IIdx));
    double p = GetProbItem(Class, Stage, Chain.GetItemIdx(IIdx, IIDS));
    //IAssert(p != 0.0);
    if (p == 0.0) { return -1e10; }
    else { return log(p); }
    //return StageProbV[Stage][Chain.GetItemIdx(IIdx, IIDS)];
  }
  double GetScoreForChain(const TChain& Chain, const int Class, const TIntV& IdxForClass, const TIntV& StagesV) {
    IAssert(IdxForClass.Len() == StagesV.Len());
    double L = 0.0;
    for (int i = 0; i < IdxForClass.Len(); i++) {
      L += GetScoreForItem(Chain, IdxForClass[i], StagesV[i], Class);
    }
    return L;
  }
  double LikelihoodTrain() {
    double L = 0.0;
    for (int c = 0; c < CIDChainH.Len(); c++) {
      TChain& Chain = CIDChainH[c];
      for (int i = 0; i < Chain.Len(); i++) {
        if (Chain.IsTest(i)) { continue; }
        L += GetScoreForItem(Chain, i, Chain.GetStage(i), Chain.GetClass(i));
      }
    }
    return L;
  }
  int GetItemsPerClassStage(TIntVV& CntVV) {
    CntVV.Gen(Classes, Stages);
    int TotalCnt = 0;
    for (int c = 0; c < CIDChainH.Len(); c++) {
      TChain& Chain = CIDChainH[c];
      int Class = Chain.GetClass();
      for (int i = 0; i < Chain.Len(); i++) {
        if (Chain.IsTest(i)) { continue; }
        int Stage = Chain.GetStage(i);
        CntVV(Class, Stage)++;
        TotalCnt++;
      }
    }
    return TotalCnt;
  }
  double EntropyAssignments() {
    double TotalEnt = 0.0;
    int TotalCnt = 0;
    TVec<TFltVVV> AlphaV, BetaV, GammaV;
    ComputeAlphaBeta(AlphaV, BetaV, GammaV);
    for (int c = 0; c < CIDChainH.Len(); c++) {
      TChain& Chain = CIDChainH[c];
      TIntV IdxV;
      for (int i = 0; i < Chain.Len(); i++) { if (! Chain.IsTest(i)) IdxV.Add(i); }
      int ChainLen = IdxV.Len();
      if (ChainLen == 0) { continue; }
      TFltVVV& Gamma = GammaV[c];
      IAssert(ChainLen == Gamma.GetZDim());
      for (int i = 0; i < ChainLen; i++) {
        double Ent = 0.0;
        for (int l = 0; l < Classes; l++) {
          for (int s = 0; s < Stages; s++) {
            Ent += exp(Gamma(l, s, i)) * Gamma(l, s, i);
          }
        }
        TotalEnt += Ent;
      }
      TotalCnt += ChainLen;
    }
    return TotalEnt / (double) TotalCnt;
  }
  double Entropy() {
    double L = 0.0;
    TFltVV EntV(Classes, Stages);
    TIntVV CntVV(Classes, Stages);
    for (int l = 0; l < Classes; l++) {
      for (int s = 0; s < Stages; s++) {
        for (int i = 0; i < Items; i++) {
          double P = GetProbItem(l, s, i);
          if (P > 0.0) {
            EntV(l, s) += - P * log (P);
          }
        }
      }
    }
    int TC = GetItemsPerClassStage(CntVV);
    for (int l = 0; l < Classes; l++) {
      for (int s = 0; s < Stages; s++) {
        L += EntV(l, s) * ((double) CntVV(l, s) / (double) TC);
      }
    }
    return L;
  }
  double ClassMetricWS(const THash<TInt, TIntSet>& IIDCatH, const int MetricType = 0) {
    //MetricType = 0: average pairwise distance, 1: - entropy
    double L = 0.0;
    TFltV ValV(Classes);
    TVec<TIntV> IdxV(Classes);
    for (int c = 0; c < CIDChainH.Len(); c++) {
      int Class = CIDChainH[c].GetClass();
      IdxV[Class].Add(c);
    }
    for (int l = 0; l < Classes; l++) {
      if (MetricType == 0) {
        int Cnt = 0;
        double Val = 0.0;
        for (int j = 0; j < IdxV[l].Len(); j++) {
          for (int k = j + 1; k < IdxV[l].Len(); k++) {
            Cnt++;
            TChain& C1 = CIDChainH[IdxV[l][j]];
            TChain& C2 = CIDChainH[IdxV[l][k]];
            Val += TChainUtil::SameCatLastItem(C1, C2, IIDCatH);
          }
        }
        if (Cnt == 0) { continue; }
        ValV[l] = Val / (double) Cnt;
      }
    }
    for (int l = 0; l < Classes; l++) {
      L += ValV[l] * ((double) IdxV[l].Len() / (double) CIDChainH.Len());
    }
    return L;
  }
  double LikelihoodTest() {
    double L = 0.0;
    for (int c = 0; c < CIDChainH.Len(); c++) {
      TChain& Chain = CIDChainH[c];
      for (int i = 0; i < Chain.Len(); i++) {
        if (! Chain.IsTest(i)) { continue; }
        //printf("Stage:%d\n", Chain.GetStageClosest(i));
        L += GetScoreForItem(Chain, i, Chain.GetStageClosest(i), Chain.GetClassClosest(i));
      }
    }
    return L;
  }
  void ClearTest() {
    for (int c = 0; c < CIDChainH.Len(); c++) {
      TChain& Chain = CIDChainH[c];
      Chain.ClearTest();
    }
  }
  void PutTestEnd(const double Frac) {
    for (int c = 0; c < CIDChainH.Len(); c++) {
      TChain& Chain = CIDChainH[c];
      Chain.ClearTest();
      int NumTest = (int) TMath::Round(Frac * (double) Chain.Len());
      if (NumTest == 0) { NumTest = 1; }
      for (int i = 0; i < NumTest; i++) {
        Chain.PutTest(Chain.Len() - 1 - i);
      }
    }
  }
  void PutTestUniform(const double Frac, int GivenSeed = -1) {
    TRnd Rnd(RndSeed);
    if (GivenSeed > 0) { Rnd.PutSeed(GivenSeed); }
    
    for (int c = 0; c < CIDChainH.Len(); c++) {
      TChain& Chain = CIDChainH[c];
      Chain.ClearTest();
      int NumTest = (int) TMath::Round(Frac * (double) Chain.Len());
      if (NumTest == 0) { NumTest = 1; }
      TIntV IdxV;
      for (int i = 0; i < Chain.Len(); i++) { IdxV.Add(i); }
      IdxV.Shuffle(Rnd);
      for (int i = 0; i < NumTest; i++) {
        Chain.PutTest(IdxV[i]);
      }
    }
  }
  void GetBestStages(const TChain& Chain, const int Class, const TIntV& IdxForClass, TIntV& NewStagesV, const bool Monotone = true) {
    int ChainLen = IdxForClass.Len();
    NewStagesV.Gen(ChainLen);
    if (ChainLen == 0) { return; }
    TFltVV BestV(ChainLen + 1, Stages + 1);
    //TIntVV DirV(ChainLen + 1, Stages + 1);
    TIntVV StagesV(ChainLen + 1, Stages + 1);
    for (int s = 0; s < Stages; s++) {
      BestV(0, s) = GetScoreForItem(Chain, IdxForClass[0], s, Class);
      //DirV(0, s) = 0;
      StagesV(0, s) = s;
    }
    for (int i = 1; i < ChainLen; i++) {
      BestV(i, 0) = GetScoreForItem(Chain, IdxForClass[i], 0, Class) + BestV(i - 1, 0);
      //DirV(i, 0) = 0;
      StagesV(i, 0) = 0;
    }
    for (int i = 1; i < ChainLen; i++) {
      for (int s = 1; s < Stages; s++) {
        double ScoreHere = GetScoreForItem(Chain, IdxForClass[i], s, Class);
        if (Monotone) {
          double Score0 = BestV(i - 1, s);
          double Score1 = BestV(i - 1, s - 1);
          if (Score0 >= Score1) {
            BestV(i, s) = ScoreHere + Score0;
            //DirV(i, s) = 0;
            StagesV(i, s) = s;
          } else {
            BestV(i, s) = ScoreHere + Score1;
            //DirV(i, s) = -1;
            StagesV(i, s) = s - 1;
          }
        } else {
          double MaxScore = TFlt::Mn;
          int MaxStage = -1;
          for (int ps = 0; ps < Stages; ps++) {
            if (MaxScore < BestV(i - 1, ps)) {
              MaxScore = BestV(i - 1, ps);
              MaxStage = ps;
            }
            BestV(i, s) = ScoreHere + MaxScore;
            StagesV(i, s) = MaxStage;
          }
        }
      }
    }
    int BestStart = -1;
    double BestScore = TFlt::Mn;
    for (int s = 0; s < Stages; s++) {
      if (BestV(ChainLen - 1, s) > BestScore) {
        double X = BestV(ChainLen - 1, s);
        BestStart = s;
        BestScore = BestV(ChainLen - 1, s);
      }
    }
    int IIdx = ChainLen - 1;
    int CurrentStage = BestStart;
    NewStagesV[IIdx] = CurrentStage;
    while (IIdx > 0) {
      //if (DirV(IIdx, CurrentStage) == -1) { CurrentStage--; }
      IIdx--;
      //NewStagesV[IIdx] = CurrentStage;
      NewStagesV[IIdx] = StagesV(IIdx + 1, NewStagesV[IIdx + 1]);
      //if (Chain.GetStage(IdxForClass[IIdx]) != CurrentStage) { Changed++; }
      //Chain.SetStage(IdxForClass[IIdx], CurrentStage);
    }
  }
  /// Given the probabilities, predict the most likely stages using dynamic programming
  int FitStage(const bool Monotone = true) {
    //int Changed = 0;
    TIntV ChangeV(CIDChainH.Len());
#pragma omp parallel for 
    for (int c = 0; c < CIDChainH.Len(); c++) {
      int Changed = 0;
      TChain& Chain = CIDChainH[c];
      //IAssert(Chain.CheckStagesIncreasing());
      TIntV IdxV;
      for (int i = 0; i < Chain.Len(); i++) { if (! Chain.IsTest(i)) IdxV.Add(i); }
      int ChainLen = IdxV.Len();
      if (ChainLen == 0) { continue; }
      TIntV BestStagesV;
      double BestScore = TFlt::Mn;
      int BestClass = -1;
      for (int Class = 0; Class < Classes; Class++) {
        TIntV CurStagesV;
        GetBestStages(Chain, Class, IdxV, CurStagesV, Monotone);
        double CurScore = GetScoreForChain(Chain, Class, IdxV, CurStagesV);
        if (CurScore > BestScore) { 
          BestScore = CurScore;
          BestStagesV = CurStagesV;
          BestClass = Class;
        }
      }
      int OldClass = Chain.GetClass();
      Chain.SetClassAll(BestClass);
      for (int i = 0; i < IdxV.Len(); i++) {
        if (OldClass == BestClass && Chain.GetStage(IdxV[i]) == BestStagesV[i].Val) { continue; }
        Chain.SetStage(IdxV[i], BestStagesV[i].Val);
        Changed++;
      }
      ChangeV[c] = Changed;
    }
    int TotalChange = 0;
    for (int c = 0; c < ChangeV.Len(); c++) { TotalChange += ChangeV[c]; }
    return TotalChange;
  }
  void Fit(const int Iter, const bool Monotone = true, const TStr LogFNm = "") {
    printf("Fitting begins\n");
    time_t InitTime = time(NULL);
    TIntV ChangeV;
    TFltV LV;
    int i = 0;
    for (i = 0; i < Iter; i++) {
      int Changes = FitStage(Monotone);
      double L = LikelihoodTrain();
      LV.Add(L);
      ChangeV.Add(Changes);
      if (Changes == 0) { break; }
      MLEProbs();
      printf("\riteration %d: %d changes[%lu secs]", i, Changes, time(NULL) - InitTime);
      fflush(stdout);
    }
    printf("MLE completed with %d iterations\n", i);
    if (LogFNm != "") {
      TGnuPlot::PlotValV(ChangeV, LogFNm + ".changes");
      TGnuPlot::PlotValV(LV, LogFNm + ".likelihood", "likelihood", "iteration", "likelihood");
    }
  }
  void FitBestL(const int Trial, const int Iter) {
    double BestL = TFlt::Mn;
    TIntV ClassV;
    TFltVVV ProbV;
    TVec<TIntV> StagesV;
    for (int s = 0; s < Trial; s++) {
      Init(s);
      Fit(Iter, "");
      double L = LikelihoodTrain();
      if (BestL < L) {
        SaveVariables(ClassV, StagesV, ProbV);
        BestL = L;
      }
    }
    LoadVariables(ClassV, StagesV, ProbV);
  }
  /// Given the probabilities, compute alpha (prob up to time t) and beta (prob from time t) using dynamic programming
  void ComputeAlphaBeta(TVec<TFltVVV>& AlphaV, TVec<TFltVVV>& BetaV, TVec<TFltVVV>& GammaV) {
    AlphaV.Gen(CIDChainH.Len());
    BetaV.Gen(CIDChainH.Len());
    GammaV.Gen(CIDChainH.Len());
#pragma omp parallel for schedule(static, 1)
    for (int c = 0; c < CIDChainH.Len(); c++) {
      TChain& Chain = CIDChainH[c];
      TIntV IdxV;
      for (int i = 0; i < Chain.Len(); i++) { if (! Chain.IsTest(i)) IdxV.Add(i); }
      int ChainLen = IdxV.Len();
      if (ChainLen == 0) { continue; }
      TFltVVV& Alpha = AlphaV[c];
      TFltVVV& Beta = BetaV[c];
      TFltVVV& Gamma = GammaV[c];
      Alpha.Gen(Classes, Stages, ChainLen);
      Beta.Gen(Classes, Stages, ChainLen);
      Gamma.Gen(Classes, Stages, ChainLen);
      for (int l = 0; l < Classes; l++) {
        //Alpha
        for (int s = 0; s < Stages; s++) {
          Alpha(l, s, 0) = log(GetProbItem(l, s, Chain.GetItemIdx(IdxV[0], IIDS)));
        }
        for (int i = 1; i < ChainLen; i++) {
          Alpha(l, 0, i) = Alpha(l, 0, i - 1) + log(GetProbItem(l, 0, Chain.GetItemIdx(IdxV[i], IIDS)));
          for (int s = 1; s < Stages; s++) {
            Alpha(l, s, i) = Alpha(l, s, i - 1) + log(1.0 + exp(Alpha(l, s - 1, i - 1) - Alpha(l, s, i - 1))) + log(GetProbItem(l, s, Chain.GetItemIdx(IdxV[i], IIDS)));
          }
        }
        //Beta
        for (int s = 0; s < Stages; s++) {
          Beta(l, s, ChainLen - 1) = log(GetProbItem(l, s, Chain.GetItemIdx(IdxV[ChainLen - 1], IIDS)));
          if (_isnan(Beta(l, s, ChainLen - 1))) {
            printf("P:%f\n", GetProbItem(l, s, Chain.GetItemIdx(IdxV[ChainLen - 1], IIDS)));
            IAssert(! _isnan(Beta(l, s, ChainLen - 1)));
          }
        }
        for (int i = ChainLen - 2; i >= 0; i--) {
          Beta(l, Stages - 1, i) = Beta(l, Stages - 1, i + 1) + log(GetProbItem(l, Stages - 1, Chain.GetItemIdx(IdxV[i], IIDS)));
          if (_isnan(Beta(l, Stages - 1, i))) {
            printf("P:%f, Prev betas:%f (s), . Two steps before: %f\n", GetProbItem(l, Stages - 1, Chain.GetItemIdx(IdxV[i], IIDS)), Beta(l, Stages - 1, i + 1).Val,
            Beta(l, Stages - 1, i + 2).Val);
            IAssert(! _isnan(Beta(l, Stages - 1, i)));
          }
          for (int s = 0; s < Stages - 1; s++) {
            //Beta(l, s, i) = log(exp(Beta(l, s, i + 1)) + exp(Beta(l, s + 1, i + 1))) + log(GetProbItem(l, s, Chain.GetItemIdx(IdxV[i], IIDS)));
            Beta(l, s, i) = Beta(l, s, i + 1) + log(1.0 + exp(Beta(l, s + 1, i + 1) - Beta(l, s, i + 1))) + log(GetProbItem(l, s, Chain.GetItemIdx(IdxV[i], IIDS)));
            if (_isnan(Beta(l, s, i))) {
              printf("P:%f, Prev betas:%f (s), %f (s + 1). Two steps before: %f, %f\n", GetProbItem(l, s, Chain.GetItemIdx(IdxV[i], IIDS)), Beta(l, s + 1, i + 1).Val, Beta(l, s, i + 1).Val,
                Beta(l, s + 1, i + 2).Val, Beta(l, s, i + 2).Val);
              IAssert(! _isnan(Beta(l, s, i)));
            }
          }
        }
        //Gamma
        for (int i = 0; i < ChainLen; i++) {
          for (int s = 0; s < Stages; s++) {
            if (GetProbItem(l, s, Chain.GetItemIdx(IdxV[i], IIDS)) == 0.0) {
              Gamma(l, s, i) = TFlt::Mn;
            } else {
              Gamma(l, s, i) = Alpha(l, s, i) + Beta(l, s, i) - log(GetProbItem(l, s, Chain.GetItemIdx(IdxV[i], IIDS)));
            }
          }
        }
      }
      //normalize gamma so that they sum up to 1
      for (int i = 0; i < ChainLen; i++) {
        double TmpVal = Gamma(0, 0, i);
        for (int l = 0; l < Classes; l++) {
          for (int s = 0; s < Stages; s++) {
            if (Gamma(l, s, i) > TmpVal) { TmpVal = Gamma(l, s, i); }
          }
        }
        for (int l = 0; l < Classes; l++) {
          for (int s = 0; s < Stages; s++) {
            Gamma(l, s, i) -= TmpVal;
          }
        }
        double SumG = 0.0;
        for (int l = 0; l < Classes; l++) {
          for (int s = 0; s < Stages; s++) {
            SumG += exp(Gamma(l, s, i));
          }
        }
        double LogSumG = log(SumG);
        for (int l = 0; l < Classes; l++) {
          for (int s = 0; s < Stages; s++) {
            Gamma(l, s, i) -= LogSumG;
          }
        }
        /*
        double Tmp = 0.0;
        for (int l = 0; l < Classes; l++) {
          for (int s = 0; s < Stages; s++) {
            Tmp += exp(Gamma(l, s, i));
          }
        }
        
        if (fabs(Tmp - 1.0) > 1e-3) {
          for (int l = 0; l < Classes; l++) {
            for (int s = 0; s < Stages; s++) {
              printf("[%d][%d] gamma:%f (%f, %f), Alpha:%f, Beta:%f, P:%f\n", l, s, Gamma(l, s, i).Val, OldGV(l, s, i).Val, OldGV2(l, s, i).Val, 
                Alpha(l, s, i).Val, Beta(l, s, i).Val, 
                GetProbItem(l, s, Chain.GetItemIdx(IdxV[i], IIDS)));
            }
          }
          IAssertR (fabs(Tmp - 1.0) < 1e-3, TStr::Fmt("sum gamma:%f, Log:%f, before norm: %f, Normval:%f\n", Tmp, LogSumG, SumG, TmpVal));
        }
        */
      }

    }
  }
  double TotalVariation(const TFltVVV& PrevP) {
    double TV = 0.0;
    for (int l = 0; l < Classes; l++) {
      for (int s = 0; s < Stages; s++) {
        for (int i = 0; i < Items; i++) {
          TV += fabs(StageProbVVV(l, s, i) - PrevP(l, s, i));
        }
      }
    }
    return TV;
  }
  void LearnEM(const int Iter, const double ThresTV, const TStr LogFNm) {
    printf("EM Fitting begins\n");
    time_t InitTime = time(NULL);
    TFltV TVV;
    TFltV LV;
    int curiter = 0;
    for (curiter = 0; curiter < Iter; curiter++) {
      // E step
      TVec<TFltVVV> AlphaV, BetaV, GammaV;
      ComputeAlphaBeta(AlphaV, BetaV, GammaV);
      // M step
      TFltVVV PrevProb = StageProbVVV;
      for (int l = 0; l < Classes; l++) {
#pragma omp parallel for
        for (int s = 0; s < Stages; s++) {
          TFltV ItemCntV(Items);
          double TotalCnt = 0.0;
          for (int c = 0; c < CIDChainH.Len(); c++) {
            TChain& Chain = CIDChainH[c];
            TIntV IdxV;
            for (int i = 0; i < Chain.Len(); i++) { if (! Chain.IsTest(i)) IdxV.Add(i); }
            int ChainLen = IdxV.Len();
            if (ChainLen == 0) { continue; }
            TFltVVV& Gamma = GammaV[c];
            IAssert(ChainLen == Gamma.GetZDim());
            for (int i = 0; i < ChainLen; i++) {
              int Item = Chain.GetItemIdx(IdxV[i], IIDS);
              IAssert(Item < Items);
              ItemCntV[Item] += exp(Gamma(l, s, i));
              TotalCnt += exp(Gamma(l, s, i));
            }
          }
          double ScaleF = 1.0;
          for (int i = 0; i < Items; i++) {
            double P = (ItemCntV[i] + ScaleF * WDirichlet) / (TotalCnt + ScaleF * WDirichlet * Items);
            if (P > 1.0) {
              printf("cnt:%f, TotalCnt:%f\n", ItemCntV[i].Val, TotalCnt);
              IAssert(P <= 1.0);
            }
            IAssert (P > 0.0);

            SetProbItem(l, s, i, P);
          }
          //printf("totalcnt:%f\n", TotalCnt);
        }
      }
      // M step end
      double TV = TotalVariation(PrevProb);
      TVV.Add(TV);
      if (TV < ThresTV) { break; }
      printf("\riteration %d: TV: %f [%lu secs]", curiter, TV, time(NULL) - InitTime);
      fflush(stdout);
    }
    printf("MLE completed with %d iterations\n", curiter);
    if (LogFNm != "") {
      TGnuPlot::PlotValV(TVV, LogFNm + ".tv", "tv", "iteration", "total variation");
    }
    FitStage();
  }
  void GetChainsByStage(THash<TInt, TIntV>& IIDCntH, const int Class) {
    IIDCntH.Gen(IIDS.Len());
    for (int i = 0; i < IIDS.Len(); i++) {
      TIntV NewV(Stages);
      IIDCntH.AddDat(IIDS[i], NewV);
    }
    for (int c = 0; c < CIDChainH.Len(); c++) {
      TChain& Chain = CIDChainH[c];
      for (int i = 0; i < Chain.Len(); i++) {
        if (Chain.IsTest(i)) { continue; }
        if (Chain.GetClass(i) != Class) { continue; }
        int Stage = Chain.GetStage(i);
        int IID = Chain.GetItemID(i);
        IIDCntH.GetDat(IID)[Stage]++;
      }
    }

  }
  void DumpStageForPatientsWithItem(const TStr& FNm, const TIntStrH& INameH, const int IID, const int Stage, const int Class = 0) {
    FILE* F = fopen(FNm.CStr(), "wt");
    for (int c = 0; c < CIDChainH.Len(); c++) {
      TChain& Chain = CIDChainH[c];
      if (! Chain.HasItemAtStage(IID, Stage, Class)) { continue; }
      fprintf(F, "Patient %d:", CIDChainH.GetKey(c).Val);
      for (int i = 0; i < Chain.Len(); i++) {
        if (Chain.IsTest(i)) { continue; }
        if (Chain.GetClass(i) != Class) { continue; }
        if (Chain.GetStage(i) != Stage) { continue; }
        int IID = Chain.GetItemID(i);
        TStr IStr = INameH.IsKey(IID)? INameH.GetDat(IID): TStr::Fmt("%d", IID);
        fprintf(F, "\t%s", IStr.CStr());
      }
      fprintf(F, "\n");
    }
    fclose(F);
  }
  void ClassByKMeans(const int Iter) {
    TVec<TIntFltH> X(CIDChainH.Len());
    TIntV MemV;
    for (int c = 0; c < CIDChainH.Len(); c++) {
      TChain& Ch = CIDChainH[c];
      for (int i = 0; i < Ch.Len(); i++) {
        if (Ch.IsTest(i)) { continue; }
        int IIdx = Ch.GetItemIdx(i, IIDS);
        if (! X[c].IsKey(IIdx)) { X[c].AddDat(IIdx, 0.0); }
        X[c].GetDat(IIdx) += 1.0;
      }
    }
    TYUtil::KMeans(X, MemV, Items, Iter, Classes, RndSeed, 0);
    for (int c = 0; c < CIDChainH.Len(); c++) {
      TChain& Ch = CIDChainH[c];
      int Class = MemV[c];
      Ch.SetClassAll(Class);
    }
  }
  void PlotBIC(const TStr& FNm, const TIntV& StagesV, const int Iter) { 
    TIntFltPrV SLV;
    ClearTest();
    int NumSamples = 0;
    for (int c = 0; c < CIDChainH.Len(); c++) {
      NumSamples += CIDChainH[c].Len();
    }
    for (int s = 0; s < StagesV.Len(); s++) {
      int CurS = StagesV[s];
      SetStages(CurS);
      Fit(Iter, "");
      double L = LikelihoodTrain();
      double BIC = 2.0 * L - (double) Classes * CurS * Items * log ((double) NumSamples);
      SLV.Add(TIntFltPr(CurS, BIC));
    }
    TGnuPlot GP(FNm);
    GP.AddPlot(SLV);
    GP.SavePng(FNm + ".png");
  }
  void PlotTopKSimPerStage(const TStr& FNm, const double TopFrac) {
    if (Classes < 2) { return; }
    TVVec<TIntSet> StageTopItemH(Classes, Stages);
    int TopK = (int) TMath::Round(TopFrac * Items);
    for (int l = 0; l < Classes; l++) {
      for (int s = 0; s < Stages; s++) {
        TIntFltH IIdxProbH;
        for (int i = 0; i < Items; i++) {
          IIdxProbH.AddDat(i, GetProbItem(l, s, i));
        }
        IIdxProbH.SortByDat(false);
        TIntSet TopItems;
        for (int i = 0; i < TopK; i++) {
          TopItems.AddKey(IIdxProbH.GetKey(i));
        }
        StageTopItemH(l, s) = TopItems;
      }
    }

    TIntFltPrV MeanV, MedianV, Q1V, Q3V;
    for (int s = 0; s < Stages; s++) {
      TMom SMom;
      for (int c1 = 0; c1 < Classes; c1++) {
        for (int c2 = 0; c2 < c1; c2++) {
          int Cap = TAGMUtil::Intersection(StageTopItemH(c1, s), StageTopItemH(c2, s));
          double Sim = (double) Cap / (double) (StageTopItemH(c1, s).Len() + StageTopItemH(c2, s).Len() - Cap);
          SMom.Add(Sim);
        }
      }
      SMom.Def();
      MeanV.Add(TIntFltPr(s + 1, SMom.GetMean()));
      MedianV.Add(TIntFltPr(s + 1, SMom.GetMedian()));
      Q1V.Add(TIntFltPr(s + 1, SMom.GetQuart1()));
      Q3V.Add(TIntFltPr(s + 1, SMom.GetQuart3()));
    }
    TGnuPlot GP(FNm);
    GP.AddPlot(MeanV, gpwLinesPoints, "Mean");
    GP.AddPlot(MedianV, gpwLinesPoints, "Median");
    GP.AddPlot(Q1V, gpwLinesPoints, "Quartile 1");
    GP.AddPlot(Q3V, gpwLinesPoints, "Quartile 3");
    GP.SavePng(FNm + ".png");
  }
  void ShuffleStage() {
    TRnd Rnd(RndSeed);
    for (int c = 0; c < CIDChainH.Len(); c++) {
      int Class = Rnd.GetUniDevInt(Classes);
      CIDChainH[c].SetClassAll(Class);
    }
  }
  void PlotAllEntPerStage(const TStr& FNm) {
    TIntFltPrV MeanV, MedianV, Q1V, Q3V, EntV;
    //TVec<TMom> SMomV;
    TIntVV CntVV;
    GetItemsPerClassStage(CntVV);
    for (int s = 0; s < Stages; s++) {
      TMom SMom;
      TFltV ProbV1(Items);
      TIntV CntV(Items);
      int TotalCnt = 0;
      for (int c = 0; c < CIDChainH.Len(); c++) {
        TChain& Chain = CIDChainH[c];
        for (int i = 0; i < Chain.Len(); i++) {
          if (Chain.IsTest(i)) { continue; }
          if (Chain.GetStage(i) != s) { continue; }
          int IIdx = Chain.GetItemIdx(i, IIDS);
          CntV[IIdx]++;
          TotalCnt++;
        }
      }
      for (int i = 0; i < Items; i++) {
        ProbV1[i] = (double) CntV[i] / (double) TotalCnt;
      }
      for (int c = 0; c < CIDChainH.Len(); c++) {
        TChain& Chain = CIDChainH[c];
        double Ent = 0.0;
        int Cnt = 0;
        for (int i = 0; i < Chain.Len(); i++) {
          if (Chain.IsTest(i)) { continue; }
          if (Chain.GetStage(i) != s) { continue; }
          int IIdx = Chain.GetItemIdx(i, IIDS);
          Ent += - log(ProbV1[IIdx]);
          Cnt++;
        }
        if (Cnt == 0) { continue; }
        Ent /= (double) Cnt;
        SMom.Add(Ent);
      }
      SMom.Def();
      MeanV.Add(TIntFltPr(s + 1, SMom.GetMean()));
      MedianV.Add(TIntFltPr(s + 1, SMom.GetMedian()));
      Q1V.Add(TIntFltPr(s + 1, SMom.GetQuart1()));
      Q3V.Add(TIntFltPr(s + 1, SMom.GetQuart3()));
      EntV.Add(TIntFltPr(s + 1, TYUtil::Entropy(ProbV1)));
    }
    TGnuPlot GP(FNm);
    GP.AddPlot(MeanV, gpwLinesPoints, "Mean");
    GP.AddPlot(MedianV, gpwLinesPoints, "Median");
    GP.AddPlot(Q1V, gpwLinesPoints, "Quartile 1");
    GP.AddPlot(Q3V, gpwLinesPoints, "Quartile 3");
    GP.AddPlot(EntV, gpwLinesPoints, "Entropy");
    GP.SavePng(FNm + ".png");
  }
  void PlotKLPerStage(const TStr& FNm, const bool Normalize = false) {
    if (Classes < 2) { return; }
  
    TStageTopics ST2(*this);
    if (Normalize) {
      ST2.Init();
    }
    TFltPrV MeanV, MedianV, Q1V, Q3V;
    TFltV VarV;
    //TVec<TMom> SMomV;
    for (int s = 0; s < Stages; s++) {
      TMom SMom;
      for (int c1 = 0; c1 < Classes; c1++) {
        TFltV ProbV1(Items);
        TFltV NPV1(Items);
        for (int i = 0; i < Items; i++) {
          ProbV1[i] = GetProbItem(c1, s, i);
          NPV1[i] = ST2.GetProbItem(c1, s, i);
        }
        for (int c2 = 0; c2 < c1; c2++) {
          TFltV ProbV2(Items);
          TFltV NPV2(Items);
          for (int i = 0; i < Items; i++) {
            ProbV2[i] = GetProbItem(c2, s, i);
            NPV2[i] = ST2.GetProbItem(c2, s, i);
          }
          double KL = TYUtil::SymKLDivergence(ProbV1, ProbV2);
          double NormKL = TYUtil::SymKLDivergence(NPV1, NPV2);
          if (Normalize) { 
            KL = KL + TYUtil::Entropy(ProbV1) + TYUtil::Entropy(ProbV2);
          }
          SMom.Add(KL);
        }
      }
      SMom.Def();
      if (Classes > 2) { VarV.Add(SMom.GetSDev()); }
      MeanV.Add(TFltPr(s + 1, SMom.GetMean()));
      MedianV.Add(TFltPr(s + 1, SMom.GetMedian()));
      Q1V.Add(TFltPr(s + 1, SMom.GetQuart1()));
      Q3V.Add(TFltPr(s + 1, SMom.GetQuart3()));
    }
    TGnuPlot GP(FNm);
    if (Classes > 2) { 
      GP.AddErrBar(MeanV, VarV, "Mean", "");
    } else {
      GP.AddPlot(MeanV, gpwLinesPoints, "Mean");
    }
    GP.AddPlot(MedianV, gpwLinesPoints, "Median");
    GP.AddPlot(Q1V, gpwLinesPoints, "Quartile 1");
    GP.AddPlot(Q3V, gpwLinesPoints, "Quartile 3");
    GP.SavePng(FNm + ".png");
  }
  void PlotCorrPerStage(const TStr& FNm) {
    if (Classes < 2) { return; }
    TStageTopics ST2(*this);
    ST2.SetStages(1);
    ST2.InitNewClasses(1);

    TIntFltPrV MeanV, MedianV, Q1V, Q3V;
    for (int s = 0; s < Stages; s++) {
      TMom SMom;
      for (int c1 = 0; c1 < Classes; c1++) {
        TFltV ProbV1(Items);
        for (int i = 0; i < Items; i++) {
          //ProbV1[i] = GetProbItem(c1, s, i) - ST2.GetProbItem(0, 0, i);
          ProbV1[i] = GetProbItem(c1, s, i) - (double) 1.0 / (double) Items;
        }
        for (int c2 = 0; c2 < c1; c2++) {
          TFltV ProbV2(Items);
          TFltPrV Prob12V;
          for (int i = 0; i < Items; i++) {
            ProbV2[i] = GetProbItem(c2, s, i) - (double) 1.0 / (double) Items;
            Prob12V.Add(TFltPr(ProbV1[i], ProbV2[i]));
          }
          double Corr = TYUtil::Corr(Prob12V);
          SMom.Add(Corr);
        }
      }
      SMom.Def();
      MeanV.Add(TIntFltPr(s + 1, SMom.GetMean()));
      MedianV.Add(TIntFltPr(s + 1, SMom.GetMedian()));
      Q1V.Add(TIntFltPr(s + 1, SMom.GetQuart1()));
      Q3V.Add(TIntFltPr(s + 1, SMom.GetQuart3()));
    }
    TGnuPlot GP(FNm);
    GP.AddPlot(MeanV, gpwLinesPoints, "Mean");
    GP.AddPlot(MedianV, gpwLinesPoints, "Median");
    GP.AddPlot(Q1V, gpwLinesPoints, "Quartile 1");
    GP.AddPlot(Q3V, gpwLinesPoints, "Quartile 3");
    GP.SavePng(FNm + ".png");
  }
  void PlotEntPerStage(const TStr& FNm) {
    TIntFltPrV MeanV, MedianV, Q1V, Q3V;
    //TVec<TMom> SMomV;
    TIntVV CntVV;
    GetItemsPerClassStage(CntVV);
    for (int s = 0; s < Stages; s++) {
      TMom SMom;
      for (int c1 = 0; c1 < Classes; c1++) {
        TFltV ProbV1(Items);
        for (int i = 0; i < Items; i++) {
          ProbV1[i] = GetProbItem(c1, s, i);
        }
        double Ent = TYUtil::Entropy(ProbV1);
        SMom.Add(Ent, (double) CntVV(c1, s));
      }
      SMom.Def();
      MeanV.Add(TIntFltPr(s + 1, SMom.GetMean()));
      MedianV.Add(TIntFltPr(s + 1, SMom.GetMedian()));
      Q1V.Add(TIntFltPr(s + 1, SMom.GetQuart1()));
      Q3V.Add(TIntFltPr(s + 1, SMom.GetQuart3()));
    }
    TGnuPlot GP(FNm);
    GP.AddPlot(MeanV, gpwLinesPoints, "Mean");
    GP.AddPlot(MedianV, gpwLinesPoints, "Median");
    GP.AddPlot(Q1V, gpwLinesPoints, "Quartile 1");
    GP.AddPlot(Q3V, gpwLinesPoints, "Quartile 3");
    GP.SavePng(FNm + ".png");
  }
  void PlotLenPerStage(const TStr& FNm) {
    TIntFltPrV MeanV, MedianV, Q1V, Q3V;
    TVec<TMom> SMomV(Stages);
    for (int c = 0; c < CIDChainH.Len(); c++) {
      TChain& Chain = CIDChainH[c];
      TIntV LenV(Stages);
      int Cnt = 0;
      for (int i = 0; i < Chain.Len(); i++) {
        if (Chain.IsTest(i)) { continue; }
        int Stage = Chain.GetStage(i);
        int IIdx = Chain.GetItemIdx(i, IIDS);
        LenV[Stage]++;
        Cnt++;
      }
      if (Cnt == 0) { continue; }
      for (int s = 0; s < Stages; s++) {
        SMomV[s].Add((double) LenV[s] / (double) Cnt);
      }
    }
    for (int s = 0; s < Stages; s++) {
      TMom& SMom = SMomV[s];
      SMom.Def();
      MeanV.Add(TIntFltPr(s + 1, SMom.GetMean()));
      MedianV.Add(TIntFltPr(s + 1, SMom.GetMedian()));
      Q1V.Add(TIntFltPr(s + 1, SMom.GetQuart1()));
      Q3V.Add(TIntFltPr(s + 1, SMom.GetQuart3()));
    }
    TGnuPlot GP(FNm);
    GP.AddPlot(MeanV, gpwLinesPoints, "Mean");
    GP.AddPlot(MedianV, gpwLinesPoints, "Median");
    GP.AddPlot(Q1V, gpwLinesPoints, "Quartile 1");
    GP.AddPlot(Q3V, gpwLinesPoints, "Quartile 3");
    GP.SavePng(FNm + ".png");
  }
  void DumpStatsPerChain(const TStr& FNm) {
    FILE* F = fopen(FNm.CStr(), "wt");
    for (int c = 0; c < CIDChainH.Len(); c++) {
      TChain& Chain = CIDChainH[c];
      TIntV LenV(Stages);
      int Cnt = 0;
      for (int i = 0; i < Chain.Len(); i++) {
        if (Chain.IsTest(i)) { continue; }
        int Stage = Chain.GetStage(i);
        int IIdx = Chain.GetItemIdx(i, IIDS);
        LenV[Stage]++;
        Cnt++;
      }
      fprintf(F, "%d\t", Chain.GetClass());
      for (int s = 0; s < Stages; s++) {
        fprintf(F, "%.4f\t", (double) LenV[s].Val / (double) Cnt);
      }
      fprintf(F, "%d\n", Cnt);
    }
    fclose(F);
  }
  void PlotFracPerStage(const TStr& FNm) {
    TIntFltPrV MeanV;
    TIntV CntV(Stages);
    for (int c = 0; c < CIDChainH.Len(); c++) {
      TChain& Chain = CIDChainH[c];
      TIntSet SS;
      for (int i = 0; i < Chain.Len(); i++) {
        if (Chain.IsTest(i)) { continue; }
        int Stage = Chain.GetStage(i);
        SS.AddKey(Stage);
      }
      for (int s = 0; s < Stages; s++) {
        if (SS.IsKey(s)) { CntV[s]++; }
      }
    }
    for (int s = 0; s < Stages; s++) {
      MeanV.Add(TIntFltPr(s + 1, (double) CntV[s] / (double) CIDChainH.Len()));
    }
    TGnuPlot GP(FNm);
    GP.AddPlot(MeanV, gpwLinesPoints, "Mean");
    GP.SavePng(FNm + ".png");
  }
  void PlotCVL(const TStr& FNm, const TIntV& StagesV, const double CVFrac, const int Trials, const int Iter) { 
    TRnd Rnd(RndSeed);
    TIntFltPrV SLV, LImpV;

    ClearTest();
    for (int s = 0; s < StagesV.Len(); s++) {
      int CurS = StagesV[s];
      SetStages(CurS);
      double L = 0.0;
      double LNull = 0.0;
      for (int t = 0; t < Trials; t++) {
        PutTestUniform(CVFrac);
        Init();
        LNull += LikelihoodTest();
        Fit(Iter, "");
        L += LikelihoodTest();
      }
      L /= (double) Trials;
      LNull /= (double) Trials;
      SLV.Add(TIntFltPr(CurS, L));
      LImpV.Add(TIntFltPr(CurS, (L - LNull)));
    }
    TGnuPlot GP(FNm);
    GP.AddPlot(SLV);
    GP.SavePng(FNm + ".png");
    TGnuPlot GP2(FNm + ".imp");
    GP2.AddPlot(LImpV);
    GP2.SavePng(FNm + ".imp.png");
  }
  void SaveVariables(TIntV& ClassV, TVec<TIntV>& StagesV, TFltVVV& ProbVVV) {
    ProbVVV = StageProbVVV;
    ClassV.Gen(CIDChainH.Len());
    StagesV.Gen(CIDChainH.Len());
    for (int c = 0; c < CIDChainH.Len(); c++) {
      TChain& Chain = CIDChainH[c];
      ClassV[c] = Chain.GetClass();
      StagesV[c].Gen(Chain.Len());
      for (int i = 0; i < Chain.Len(); i++) {
        if (Chain.IsTest(i)) { continue; }
        StagesV[c][i] = Chain.GetStage(i);
      }
    }
  }
  void LoadVariables(const TIntV& ClassV, const TVec<TIntV>& StagesV, const TFltVVV& ProbVVV) {
    StageProbVVV = ProbVVV;
    IAssert(ClassV.Len() == CIDChainH.Len());
    for (int c = 0; c < CIDChainH.Len(); c++) {
      TChain& Chain = CIDChainH[c];
      Chain.SetClassAll(ClassV[c]);
      for (int i = 0; i < Chain.Len(); i++) {
        if (Chain.IsTest(i)) { continue; }
        Chain.SetStage(i, StagesV[c][i]);
      }
    }
  }
  void GetChainClasses(TIntV& CCntV) {
    CCntV.Gen(Classes);
    for (int c = 0; c < CIDChainH.Len(); c++) {
      if (CIDChainH[c].GetClass() < 0) { continue; }
      CCntV[CIDChainH[c].GetClass()]++;
    }
  }
  double TimeAtFraction(const double Frac) {
    // find time at (1 - Fraction) of data
    TFltV TimeV;
    for (int c = 0; c < CIDChainH.Len(); c++) {
      for (int i = 0; i < CIDChainH[c].Len(); i++) {
        if (CIDChainH[c].IsTest(i)) { continue; }
        TimeV.Add(CIDChainH[c].GetTime(i));
      }
    }
    TimeV.Sort(false);
    const int ItemsToCut = (int) TMath::Round(Frac * (double) TimeV.Len());
    return TimeV[ItemsToCut].Val;
  }
  void EstimateTopItemsForTestSet(TIntSet& TopIIDS, const int TopK) {
    TIntFltH IIDCntH;
    for (int c = 0; c < CIDChainH.Len(); c++) {
      TChain& Chain = CIDChainH[c];
      for (int i = 0; i < Chain.Len(); i++) {
        if (! Chain.IsTest(i)) { continue; }
        int IID = Chain.GetItemID(i);
        int IIdx = Chain.GetItemIdx(i, IIDS);
        IIDCntH.AddDat(IID) += GetProbItem(Chain.GetClassClosest(i), Chain.GetStageClosest(i), IIdx);
      }
    }
    IIDCntH.SortByDat(false);
    TopIIDS.Gen(TopK);
    for (int i = 0; i < TopK; i++) {
      int IID = IIDCntH.GetKey(i);
      TopIIDS.AddKey(IID);
    }
  }

  void GTTopItemsForTestSet(TIntSet& TopIIDS, const int TopK) {
    TIntH IIDCntH;
    for (int c = 0; c < CIDChainH.Len(); c++) {
      TChain& Chain = CIDChainH[c];
      for (int i = 0; i < Chain.Len(); i++) {
        if (! Chain.IsTest(i)) { continue; }
        int IID = Chain.GetItemID(i);
        IIDCntH.AddDat(IID)++;
      }
    }
    IIDCntH.SortByDat(false);
    TopIIDS.Gen(TopK);
    for (int i = 0; i < TopK; i++) {
      int IID = IIDCntH.GetKey(i);
      TopIIDS.AddKey(IID);
    }
  }
  void FindTopItemsForNextAbsoluteTime(TIntSet& TopIIDS, const int TopK, const double FracForCnt = 0.1) {
    double TimeForCnt = TimeAtFraction(FracForCnt);
    TIntFltH IIdxCntH;
    TIntVV CntVV(Classes, Stages);
    for (int c = 0; c < CIDChainH.Len(); c++) {
      TChain& Chain = CIDChainH[c];
      int Class = Chain.GetClass();
      int Cnt = 0;
      int LIdx = Chain.Len() - 1;
      while (Chain.IsTest(LIdx)) { LIdx--; }
      int LStage = Chain.GetStage(LIdx);
      for (int i = Chain.Len() - 1; i >= 0; i--) {
        if (Chain.IsTest(i)) { continue; }
        if (Chain.GetTime(i) < TimeForCnt) { break; }
        Cnt++;
      }
      if (Cnt == 0) { continue; }
      CntVV(Class, LStage) += Cnt;
    }
    //get rate * probability for each item
    for (int l = 0; l < Classes; l++) {
      for (int s = 0; s < Stages; s++) {
        for (int i = 0; i < Items; i++) {
          IIdxCntH.AddDat(i) += GetProbItem(l, s, i) * (double) CntVV(l, s);
        }
      }
    }
    IIdxCntH.SortByDat(false);
    TopIIDS.Gen(TopK);
    for (int i = 0; i < TopK; i++) {
      int IID = IIDS.GetKey(IIdxCntH.GetKey(i));
      TopIIDS.AddKey(IID);
    }
  }
};


#endif