#ifndef yang_utils_h
#define yang_utils_h

#include "Snap.h"

class TYUtil{
public:
  TYUtil(void) {}
  ~TYUtil(void) {}
  static PUNGraph LoadGML(const TStr& InFNm);
  static void GetPDFForBins(TFltFltH& ValPDFH, const TFltV& BinValV, const TFltV& PointsV);
  static void GetPDFForBinsFloor(TFltFltH& ValPDFH, const TFltV& BinValV, const TFltV& PointsV);
  static void GetCCDF(const TFltV& ValV, TFltFltH& CCDFH, const bool CCDF);
  static void GetCCDFFromPDF(THash<TFlt, TInt>& PDFH, TFltFltH& CCDFH, const bool CCDF);
  static TStrV GetFileNames(const TStr DirStr) {
    PFFile Files = TFFile::New(DirStr, false);
    TStr TmpFNm;
    TStrV FNmV;
    while (Files->Next(TmpFNm)) { // get files in the data path
      FNmV.Add(TmpFNm);
    }
    return FNmV;
  }
  //static int GetNIDWithMinFarness(const PUNGraph& SG, const int MxSzFullSearch);
  static void GetCCDFFromPDF(TIntH& PDFH, TIntFltH& CCDFH, const bool CCDF = true); 
  static double Norm(const TVec<TFltV>& X);
  static double F1Score(const TIntSet& GT, const TIntSet& Est, double& Pr, double& Rc); 
  static double F1Score(const TIntSet& GT, const TIntSet& Est) {
    double Pr, Rc;
    return F1Score(GT, Est, Pr, Rc);
  }

  static void NormalizeV(TFltPrV& ValueV, const double MaxV);
  static double Corr(TFltPrV& XYV) {
    double EX = 0.0, EY = 0.0, EX2 = 0.0, EY2 = 0.0;
    double EXY = 0.0;
    TMom YM, XM, XYM;
    for (int i = 0; i < XYV.Len(); i++) {
      XM.Add(XYV[i].Val1);
      YM.Add(XYV[i].Val2);
      XYM.Add(XYV[i].Val1 * XYV[i].Val2);
    }
    XM.Def(); YM.Def(); XYM.Def();
    return (XYM.GetMean() - XM.GetMean() * YM.GetMean()) / XM.GetSDev() / YM.GetSDev();
  }
  static double RankCorr(TFltPrV& XYV) {
    int NoMatch = 0;
    int Cnt = 0;
    for (int i = 1; i < XYV.Len(); i++) {
      for (int j = 0; j < i; j++) {
        if (XYV[i].Val1 < XYV[j].Val1 && XYV[i].Val2 > XYV[j].Val2) {
          NoMatch++;
        } else if (XYV[i].Val1 > XYV[j].Val1 && XYV[i].Val2 < XYV[j].Val2) {
          NoMatch++;
        }
        Cnt++;
      }
    }
    return (double) (Cnt - NoMatch) / (double) Cnt;
  }
  static void LibLinearTrain(const TStr DataFNm, const TStr ModelFNm, const TStr LibPath, const TVec<TIntFltH>& XV, const TIntV& YV, const TStr OptionStr) {
    IAssert(XV.Len() == YV.Len());
    FILE* F = fopen(DataFNm.CStr(), "wt");
    for (int i = 0; i < XV.Len(); i++) {
      fprintf(F, "%d", YV[i].Val);
      for (int x = 0; x < XV[i].Len(); x++) {
        fprintf(F, " %d:%f", XV[i].GetKey(x).Val + 1, XV[i][x].Val);
      }
      fprintf(F, "\n");
    }
    fclose(F);
    TStr CmdStr = TStr::Fmt("%s %s -q %s %s", LibPath.CStr(), OptionStr.CStr(), DataFNm.CStr(), ModelFNm.CStr());
    int Res = system(CmdStr.CStr());

  }
  static void LibLinearTest(const TStr DataFNm, const TStr ModelFNm, const TStr OutFNm, const TStr LibPath, const TVec<TIntFltH>& XV, const TIntV& YV, TVec<TIntFltH>& ProbHV, const TStr OptionStr) {
    IAssert(XV.Len() == YV.Len());
    FILE* F = fopen(DataFNm.CStr(), "wt");
    for (int i = 0; i < XV.Len(); i++) {
      fprintf(F, "%d", YV[i].Val);
      for (int x = 0; x < XV[i].Len(); x++) {
        fprintf(F, " %d:%f", XV[i].GetKey(x).Val + 1, XV[i][x].Val);
      }
      fprintf(F, "\n");
    }
    fclose(F);
    TStr CmdStr = TStr::Fmt("%s %s %s %s %s", LibPath.CStr(), OptionStr.CStr(), DataFNm.CStr(), ModelFNm.CStr(), OutFNm.CStr());
    int Res = system(CmdStr.CStr());
    //read output
    TSsParser Ss(OutFNm, ssfWhiteSep);
    Ss.Next();
    int Labels = Ss.GetFlds() - 1;
    TIntV LabelV;
    for (int i = 1; i < Ss.GetFlds(); i++) { LabelV.Add(Ss.GetInt(i)); }
    ProbHV.Clr();
    int LCnt = 0;
    while (Ss.Next()) {
      TIntFltH ProbH(Ss.GetFlds() - 1);
      for (int i = 1; i < Ss.GetFlds(); i++) { ProbH.AddDat(LabelV[i - 1], Ss.GetFlt(i)); }
      ProbH.SortByDat(false);
      ProbHV.Add(ProbH);
    }
    IAssert(YV.Len() == ProbHV.Len());
  }
  static double MeasureMultLikelihood(const TIntV& YV, const TVec<TIntFltH>& ProbHV, const double BaseP) {
    double L = 0.0;
    IAssert (ProbHV.Len() == YV.Len());
    for (int i = 0; i < ProbHV.Len(); i++) {
      double P = BaseP;
      if (ProbHV[i].IsKey(YV[i])) {
        P = ProbHV[i].GetDat(YV[i].Val);
      }
      L += log (P);
    }
    return L;
  }
  static double MeasureAccTopK(const TIntV& YV, const TVec<TIntFltH>& ProbHV, const int TopK) {
    int Hits = 0;
    int Cnt = 0;
    IAssert (ProbHV.Len() == YV.Len());
    for (int i = 0; i < ProbHV.Len(); i++) {
      if (ProbHV[i].IsKey(YV[i]) && ProbHV[i].GetKeyId(YV[i]) < TopK) { Hits++; }
      Cnt++;
    }
    return (double) Hits / (double) Cnt;

  }
  static void ReadStrStrH(const TStr& FNm, TStrStrH& OutH, TSsFmt Sep);
  static void ReadStrHashWithID(const TStr& FNm, TStrHash<TInt>& OutH, TSsFmt Sep, const int StrColumn = 0, const int IntColumn = 1);
  static void ReadIntStrH(const TStr& FNm, TIntStrH& OutH, TSsFmt Sep);
  static void ReadStrHash(const TStr& FNm, TStrHash<TInt>& OutH);
  static void ReadIntH(const TStr& FNm, TIntH& OutH, TSsFmt Sep);
  static void ReadIntFltH(const TStr& FNm, TIntFltH& OutH, TSsFmt Sep) {
    TSsParser Ss(FNm, Sep);
    while (Ss.Next()) {
      if (Ss.Len() < 2) { continue; }
      OutH.AddDat(Ss.GetInt(0), Ss.GetFlt(1));
    }
  }
  static void ReadIntSet(const TStr& FNm, TIntSet& OutH) {
    TSsParser Ss(FNm);
    while (Ss.Next()) {
      if (Ss.Len() < 1) { continue; }
      OutH.AddKey(Ss.GetInt(0));
    }
  }
  static void ReadStrSet(const TStr& FNm, TStrSet& OutH) {
    TSsParser Ss(FNm);
    while (Ss.Next()) {
      if (Ss.Len() < 1) { continue; }
      OutH.AddKey(Ss.GetFld(0));
    }
  }
  static void DumpIntStrH(const TStr& FNm, TIntStrH& H, const TStr Sep) {
    FILE* F = fopen(FNm.CStr(), "wt");
    for (int i = 0; i < H.Len(); i++) {
      fprintf(F, "%d%s%s\n", H.GetKey(i).Val, Sep.CStr(), H[i].CStr());
    }
    fclose(F);
  }

  static void GetKMeansCent(const TVec<TIntFltH>& X, const TIntV& MemV, const int XCols, TVec<TFltV>& Cents) {
    IAssert(MemV.Len() == X.Len());
    int Classes = Cents.Len();
    TIntV CntV(Classes);
    for (int l = 0; l < Classes; l++) {
      Cents[l].Gen(XCols);
    }
    for (int c = 0; c < X.Len(); c++) {
      int Class = MemV[c];
      CntV[Class]++;
      for (int i = 0; i < X[c].Len(); i++) {
        int Idx = X[c].GetKey(i);
        double Val = X[c][i];
        Cents[Class][Idx] += Val;
      }
    }
    for (int l = 0; l < Classes; l++) {
      if (CntV[l] == 0) { continue; }
      for (int i = 0; i < XCols; i++) { Cents[l][i] /= CntV[l]; }
    }
  }
  static double KLDivergence(const TFltV& X, const TFltV& Y) {
    IAssert(X.Len() == Y.Len());
    double L = 0.0;
    for (int i = 0; i < X.Len(); i++) {
      if (X[i] > 0.0 && Y[i] > 0.0) {
        L += log (X[i].Val / Y[i].Val) * X[i];
      }
    }
    return L;
  }
  static double Entropy(const TFltV& X) {
    double L = 0.0;
    for (int i = 0; i < X.Len(); i++) {
      IAssert(X[i] >= 0.0);
      if (X[i] > 0.0) {
        L += - X[i] * log(X[i].Val);
      }
    }
    return L;
  }
  static double SymKLDivergence(const TFltV& X, const TFltV& Y) {
    return KLDivergence(X, Y) + KLDivergence (Y, X);
  }
  static void GetLinIntV(const int Min, const int Max, const int Divs, TIntV& IntV) {
    IAssert(Divs > 1);
    IntV.Clr();
    double Interval = (double) (Max - Min) / (double) (Divs - 1);
    for (int i = 0; i < Divs; i++) {
      IntV.Add((int) TMath::Round((double) Min + (double) i * Interval));
    }
    
  }
  static void GetLogIntV(const int Min, const int Max, const int Divs, TIntV& IntV) {
    IAssert(Divs > 1);
    IntV.Clr();
    double Interval = (log((double) Max) - log((double) Min)) / (double) (Divs - 1);
    for (int i = 0; i < Divs; i++) {
      IntV.Add((int) TMath::Round((double) Min * exp(Interval * i)));
    }
  }
  static double Norm(const TIntFltH& X) {
    double L = 0.0;
    for (int i = 0; i < X.Len(); i++) { L += X[i].Val * X[i].Val; }
    return sqrt(L);
  }
  static double DotProduct(const TIntFltH& X, const TFltV& Cent) {
    double L = 0.0;
    for (int i = 0; i < X.Len(); i++) { L += X[i] * Cent[X.GetKey(i)]; }
    return L;

  }
  static double CosineDist(const TIntFltH& X, const TFltV& Cent) {
    double Norm1 = Norm(X);
    double Norm2 = TLinAlg::Norm(Cent);
    if (Norm1 == 0.0 || Norm2 == 0.0) { return TFlt::Mx; }
    else {
      return 1.0 - DotProduct(X, Cent) / Norm1 / Norm2;
    }
  }
  static void KMeans(const TVec<TIntFltH>& X, TIntV& MemV, const int XCols, const int Iter, const int Classes, const int RndSeed = 0, const int MetricType = 0) {
    //MetricType == 0: 1 - cosine sim, 1: Euclidean distance
    TRnd Rnd(RndSeed);
    TVec<TFltV> Cents(Classes);
    MemV.Gen(X.Len());
    //initialize membership randomly
    for (int i = 0; i < X.Len(); i++) {
      MemV[i] = Rnd.GetUniDevInt(Classes);
    }
    TIntV PrevMemV;
    int it = 0;
    for (it = 0; it < Iter; it++) {
      int Changes = 0;
      GetKMeansCent(X, MemV, XCols, Cents);
#pragma omp parallel for
      for (int i = 0; i < X.Len(); i++) {
        double MinDist = TFlt::Mx;
        int BestC = -1;
        for (int c = 0; c < Classes; c++) {
          double D = CosineDist(X[i], Cents[c]);
          if (D < MinDist) { 
            BestC = c;
            MinDist = D;
          }
        }
        if (BestC != MemV[i]) {
#pragma omp atomic
          Changes++;
        }
        MemV[i] = BestC;
      }
      printf("\r%d iteration [%d changes]", it, Changes);
      if (Changes == 0) { break; }
    }
    printf("\nK-Means done with %d iterations\n", it + 1);
  }


  template <class PGraph> void SavePajek(const PGraph& Graph, const TStr& OutFNm, const TStr& DescStr, const int SampleNodes = -1);
//--------------------------------------------------------------------------------
//Implementation
  template <class PGraph>
  static PGraph GetIntersectionGraph(const PGraph& G1, const PGraph& G2) {
    PGraph Graph = PGraph::TObj::New();
    if (G1->GetNodes() < G2->GetNodes()) {
      for (typename PGraph::TObj::TNodeI NI = G1->BegNI(); NI < G1->EndNI(); NI++) {
        if (G2->IsNode(NI.GetId())) { Graph->AddNode(NI.GetId()); }
      }
    } else {
      for (typename PGraph::TObj::TNodeI NI = G2->BegNI(); NI < G2->EndNI(); NI++) {
        if (G1->IsNode(NI.GetId())) { Graph->AddNode(NI.GetId()); }
      }
    }
    if (G1->GetEdges() < G2->GetEdges()) {
      for (typename PGraph::TObj::TEdgeI EI = G1->BegEI(); EI < G1->EndEI(); EI++) {
        const int SrcNId = EI.GetSrcNId();
        const int DstNId = EI.GetDstNId();
        if (G2->IsEdge(SrcNId, DstNId)) { Graph->AddEdge(SrcNId, DstNId); }
      }
    } else {
      for (typename PGraph::TObj::TEdgeI EI = G2->BegEI(); EI < G2->EndEI(); EI++) {
        const int SrcNId = EI.GetSrcNId();
        const int DstNId = EI.GetDstNId();
        if (G1->IsEdge(SrcNId, DstNId)) { Graph->AddEdge(SrcNId, DstNId); }
      }
    }
    return Graph;
  }

  template <class PGraph>
  static PGraph LoadEdgeListStr(const TStr& InFNm, TIntStrH& NIDNameH, const int& SrcColId = 0, const int& DstColId = 1, const TSsFmt SsFmt = ssfTabSep) {
    TSsParser Ss(InFNm, SsFmt);
    PGraph Graph = PGraph::TObj::New();
    TStrHash<TInt> StrSet(Mega(1), true);
    while (Ss.Next()) {
      const int SrcNId = StrSet.AddKey(Ss[SrcColId]);
      const int DstNId = StrSet.AddKey(Ss[DstColId]);
      if (! Graph->IsNode(SrcNId)) { Graph->AddNode(SrcNId); }
      if (! Graph->IsNode(DstNId)) { Graph->AddNode(DstNId); }
      Graph->AddEdge(SrcNId, DstNId);
    }
    NIDNameH.Gen(StrSet.Len());
    for (int s = 0; s < StrSet.Len(); s++) { NIDNameH.AddDat(s, StrSet.GetKey(s)); }
    IAssert(NIDNameH.Len() == Graph->GetNodes());

    Graph->Defrag();
    return Graph;
  }

  template <class PGraph> 
  static void SavePajek(const PGraph& Graph, const TStr& OutFNm, const TStr& DescStr) {
    TIntH NIdToIdH(Graph->GetNodes(), true);
    FILE *F = fopen(OutFNm.CStr(), "wt");
    fprintf(F, "*Vertices %d\n", Graph->GetNodes());
    int i = 0;
    for (typename PGraph::TObj::TNodeI NI = Graph->BegNI(); NI < Graph->EndNI(); NI++, i++) {
      if (i == 0) {
        fprintf(F, "%d  \"%s\" ic Red fos 10\n", i+1, DescStr.CStr()); // ic: internal color, fos: font size
      } else {
        fprintf(F, "%d  \"\" ic Red fos 10\n", i+1); // ic: internal color, fos: font size
      }
      NIdToIdH.AddDat(NI.GetId(), i+1);
    }
    if (HasGraphFlag(typename PGraph::TObj, gfDirected)) {
      fprintf(F, "*Arcs %d\n", Graph->GetEdges()); } // arcs are directed, edges are undirected
    else {
      fprintf(F, "*Edges %d\n", Graph->GetEdges());
    }
    for (typename PGraph::TObj::TEdgeI EI = Graph->BegEI(); EI < Graph->EndEI(); EI++) {
      const int SrcNId = NIdToIdH.GetDat(EI.GetSrcNId());
      const int DstNId = NIdToIdH.GetDat(EI.GetDstNId());
      fprintf(F, "%d %d %d c Black\n", SrcNId, DstNId, 1);
    }
    fclose(F);
    
  }

  template <class PGraph>
  static void PlotClustCfCnt(const PGraph& Graph, const TStr& FNmPref, TStr DescStr, const int SampleNodes) {
    THash<TFlt, TInt> CCFCntH(SampleNodes / 5);
    const double CCF = GetClustCfCnt(Graph, CCFCntH, SampleNodes);
    if (DescStr.Empty()) { DescStr = FNmPref; }
    TGnuPlot GnuPlot("ccfcnt."+FNmPref);
    GnuPlot.SetTitle(TStr::Fmt("Average Clustering coefficient: %f, %s", CCF, DescStr.CStr()));
    GnuPlot.AddPlot(CCFCntH, gpwLinesPoints, "", "pt 6");
    GnuPlot.SetXYLabel("CCF", "Counts");
    GnuPlot.SetScale(gpsLog10XY);
    GnuPlot.SavePng();
  }

  template <class PGraph> static double GetClustCfCnt(const PGraph& Graph, THash<TFlt, TInt>& CCFCntH, const int SampleNodes) {
    TIntTrV NIdCOTriadV;
    TSnap::GetTriads(Graph, NIdCOTriadV, SampleNodes);
    THash<TInt, TFltPr> DegSumCnt;
    double SumCcf = 0.0;
    for (int i = 0; i < NIdCOTriadV.Len(); i++) {
      const int D = NIdCOTriadV[i].Val2()+NIdCOTriadV[i].Val3();
      const double Ccf = D!=0 ? NIdCOTriadV[i].Val2() / double(D) : 0.0;
      SumCcf += Ccf;
      CCFCntH.AddDat(Ccf)++;
    }
    CCFCntH.SortByDat(true);
    return SumCcf / double(NIdCOTriadV.Len());
  }

  template <class PGraph>
  static PGraph GetEgoNet(const PGraph& Graph, const int CentCID, const int Hops, const bool TakeWcc = false, const bool FollowOut = true, const bool FollowIn = true) {
    TIntV FocusNIdV;
    TBreathFS<PGraph> BFS(Graph);
    BFS.DoBfs(CentCID, FollowOut, FollowIn);
    BFS.NIdDistH.SortByDat(true);
    for (int d = 0; d < BFS.NIdDistH.Len(); d++) {
      if (BFS.NIdDistH[d] > Hops) { break; }
      if (BFS.NIdDistH[d] == 0) { continue; }
      FocusNIdV.Add(BFS.NIdDistH.GetKey(d));
    }
    PGraph TmpUG = TSnap::GetSubGraph<PGraph>(Graph, FocusNIdV);
    if (TakeWcc) {
      return TSnap::GetMxWcc(TmpUG);
    } else {
      return TmpUG;
    }
  }


};

class TAGMUtil{
public:
  static void GenPLSeq(TIntV& SzSeq,const int& SeqLen, const double& Alpha, TRnd& Rnd, const int& Min, const int& Max);
  static void ConnectCmtyVV(TVec<TIntV>& CmtyVV, const TIntPrV& CIDSzPrV, const TIntPrV& NIDMemPrV, TRnd& Rnd);
  static void GenCmtyVVFromPL(TVec<TIntV>& CmtyVV, const PUNGraph& Graph, const int& Nodes, const int& Coms, const double& ComSzAlpha, const double& MemAlpha, const int& MinSz, const int& MaxSz, const int& MinK, const int& MaxK, TRnd& Rnd);
  static void GenCmtyVVFromPL(TVec<TIntV>& CmtyVV, const TIntV& NIDV, const int& Nodes, const int& Coms, const double& ComSzAlpha, const double& MemAlpha, const int& MinSz, const int& MaxSz, const int& MinK, const int& MaxK, TRnd& Rnd);
  static void GenCmtyVVFromPL(TVec<TIntV>& CmtyVV, const int& Nodes, const int& Coms, const double& ComSzAlpha, const double& MemAlpha, const int& MinSz, const int& MaxSz, const int& MinK, const int& MaxK, TRnd& Rnd);
  static void RndConnectInsideCommunity(PUNGraph& Graph, const TIntV& CmtyV, const double& Prob, TRnd& Rnd);
  static PUNGraph GenAGM(TVec<TIntV >& CmtyVV, const double& DensityCoef, const double& ScaleCoef, TRnd& Rnd=TInt::Rnd);
  static PUNGraph GenAGM(TVec<TIntV>& CmtyVV, const double& DensityCoef, const int TargetEdges, TRnd& Rnd);
  static PUNGraph GenAGM(TVec<TIntV>& CmtyVV, const TFltV& CProbV, TRnd& Rnd, const double PNoCom = -1.0);
  static void GetNodeMembership(THash<TInt,TIntSet >& NIDComVH, const TVec<TIntV>& CmtyVV);
  static void GetNodeMembership(THash<TInt,TIntSet >& NIDComVH, const TVec<TIntV>& CmtyVV, const TIntV& NIDV);
  static void GetNodeMembership(TIntH& NIDComVH, const THash<TInt,TIntV >& CmtyVH);
  static void GetNodeMembership(THash<TInt,TIntSet >& NIDComVH, const TVec<TIntSet>& CmtyVV);
  static void GetNodeMembership(THash<TInt,TIntSet >& NIDComVH, const THash<TInt,TIntV >& CmtyVH);
  static void GetNodeMembership(THash<TInt,TIntV >& NIDComVH, const THash<TInt,TIntV >& CmtyVH);
  static void GetNodeMembership(THash<TInt,TIntV >& NIDComVH, const TVec<TIntV >& CmtyVV);
  static void GetNodeMembership(TIntH& NIDComVH, const TVec<TIntV >& CmtyVV);
  static void LoadCmtyVV(const TStr& InFNm, TVec<TIntV>& CmtyVV, const int BeginCol = 0, const int MinSz = 3, const TSsFmt Sep = ssfWhiteSep);
  static void LoadCmtyVV(const TStr& InFNm, TVec<TIntV>& CmtyVV, TStrHash<TInt>& StrToNIdH, const int BeginCol = 0, const int MinSz = 3, const TSsFmt Sep = ssfWhiteSep);
  static void LoadCmtyVH(const TStr& InFNm, THash<TInt, TIntV>& CmtyVH, const int MinSz = 3, const TSsFmt Sep = ssfWhiteSep);
  static void DumpCmtyVV(const TStr& OutFNm, const TVec<TIntV>& CmtyVV);
  static void DumpCmtyVV(const TStr OutFNm, TVec<TIntV>& CmtyVV, TStrHash<TInt>& NIDNmH);
  static void DumpCmtyVV(const TStr OutFNm, TVec<TIntV>& CmtyVV, TIntStrH& NIDNmH);
  static void CmtyVVContainingNID(const TVec<TIntV>& CmtyVVIn, TVec<TIntV>& CmtyVVOut, const int NId);
  static int TotalMemberships(const TVec<TIntV>& CmtyVV);
  static void RewireCmtyVV(const TVec<TIntV>& CmtyVVIn, TVec<TIntV>& CmtyVVOut, TRnd& Rnd);
  static void RewireCmtyNID(THash<TInt,TIntV >& CmtyVH, TRnd& Rnd);
  static int Intersection(const TIntV& C1, const TIntV& C2);
  static void GetIntersection(const THashSet<TInt>& A, const THashSet<TInt>& B, THashSet<TInt>& C);
  static int Intersection(const THashSet<TInt>& A, const THashSet<TInt>& B);
  //static double GetConductance(const PUNGraph& Graph, const TIntSet& CmtyS, const int Edges);
  //static double GetConductance(const PNGraph& Graph, const TIntSet& CmtyS, const int Edges);
  template<class PGraph>
  static double GetConductance(const PGraph& Graph, const TIntSet& CmtyS, const int Edges) {
    const bool GraphType = HasGraphFlag(typename PGraph::TObj, gfDirected);
    int Edges2;
    if (GraphType) { Edges2 = Edges >= 0 ? Edges : Graph->GetEdges(); }
    else { Edges2 = Edges >= 0 ? 2 * Edges : Graph->GetEdges(); }
    int Vol = 0,  Cut = 0; 
    double Phi = 0.0;
    for (int i = 0; i < CmtyS.Len(); i++) {
      if (! Graph->IsNode(CmtyS[i])) { continue; }
      typename PGraph::TObj::TNodeI  NI = Graph->GetNI(CmtyS[i]);
      for (int e = 0; e < NI.GetOutDeg(); e++) {
        if (! CmtyS.IsKey(NI.GetOutNId(e))) { Cut += 1; }
      }
      Vol += NI.GetOutDeg();
    }
    // get conductance
    if (Vol != Edges2) {
      if (2 * Vol > Edges2) { Phi = Cut / double (Edges2 - Vol); }
      else if (Vol == 0) { Phi = 0.0; }
      else { Phi = Cut / double(Vol); }
    } else {
      if (Vol == Edges2) { Phi = 1.0; }
    }
    return Phi;
  }

  static void GetNbhCom(const PUNGraph& Graph, const int NID, TIntSet& NBCmtyS);
  static void GetNbhCom(const PNGraph& Graph, const int NID, TIntSet& NBCmtyS);
  template<class PGraph>
  static void SaveGephi(const TStr& OutFNm, const PGraph& G, const TVec<TIntV >& CmtyVVAtr) {
    THash<TInt, TStr> TmpH;
    SaveGephi(OutFNm, G, CmtyVVAtr, TmpH);
  }
  template<class PGraph>
  static void SaveGephi(const TStr& OutFNm, const PGraph& G) {
    TVec<TIntV> CmtyVVAtr;
    SaveGephi(OutFNm, G, CmtyVVAtr);
  }
  template<class PGraph>
  static void SaveGephi(const TStr& OutFNm, const PGraph& G, const TVec<TIntV >& CmtyVVAtr, const THash<TInt, TStr>& NIDNameH) { 
    THash<TInt, TIntTr> TmpH1; 
    TIntFltH TmpH2;
    THash<TIntPr, TIntTr> TmpH3;
    TIntPrFltH TmpH4;
    SaveGephi(OutFNm, G, CmtyVVAtr, NIDNameH, TmpH1, TmpH2, TmpH3, TmpH4);
  }
  template<class PGraph>
  static void SaveGephi(const TStr& OutFNm, const PGraph& G, const TVec<TIntV >& CmtyVVAtr, const THash<TInt, TStr>& NIDNameH, const THash<TInt, TIntTr >& NIDColorH) { 
    TIntFltH TmpH2;
    THash<TIntPr, TIntTr> TmpH3;
    TIntPrFltH TmpH4;
    SaveGephi(OutFNm, G, CmtyVVAtr, NIDNameH, NIDColorH, TmpH2, TmpH3, TmpH4);
  }
  template<class PGraph>
  static void SaveGephi(const TStr& OutFNm, const PGraph& G, const TVec<TIntV >& CmtyVVAtr, const THash<TInt, TStr>& NIDNameH, const THash<TInt, TIntTr >& NIDColorH, const TIntFltH& NIDSizeH, const THash<TIntPr, TIntTr >& EIDColorH, const TIntPrFltH& EIDSizeH) {
    const bool GraphType = HasGraphFlag(typename PGraph::TObj, gfDirected);
    THash<TInt,TIntV> NIDComVHAtr;
    TAGMUtil::GetNodeMembership(NIDComVHAtr, CmtyVVAtr);

    FILE* F = fopen(OutFNm.CStr(), "wt");
    fprintf(F, "<?xml version='1.0' encoding='UTF-8'?>\n");
    fprintf(F, "<gexf xmlns='http://www.gexf.net/1.2draft' xmlns:viz='http://www.gexf.net/1.1draft/viz' xmlns:xsi='http://www.w3.org/2001/XMLSchema-instance' xsi:schemaLocation='http://www.gexf.net/1.2draft http://www.gexf.net/1.2draft/gexf.xsd' version='1.2'>\n");
    if (GraphType) {
      fprintf(F, "\t<graph mode='static' defaultedgetype='directed'>\n");
    } else {
      fprintf(F, "\t<graph mode='static' defaultedgetype='undirected'>\n");
    }
    if (CmtyVVAtr.Len() > 0) {
      fprintf(F, "\t<attributes class='node'>\n");
      for (int c = 0; c < CmtyVVAtr.Len(); c++) {
        fprintf(F, "\t\t<attribute id='%d' title='c%d' type='boolean'>", c, c);
        fprintf(F, "\t\t<default>false</default>\n");
        fprintf(F, "\t\t</attribute>\n");
      }
      fprintf(F, "\t</attributes>\n");
    }
    fprintf(F, "\t\t<nodes>\n");
    for (typename PGraph::TObj::TNodeI NI = G->BegNI(); NI < G->EndNI(); NI++) {
      int NID = NI.GetId();
      TStr Label = NIDNameH.IsKey(NID)? NIDNameH.GetDat(NID): TStr::Fmt("%d", NID);
      TIntTr Color = NIDColorH.IsKey(NID)? NIDColorH.GetDat(NID) : TIntTr(120, 120, 120);
      Label.ChangeChAll('<', ' ');
      Label.ChangeChAll('>', ' ');
      Label.ChangeChAll('&', ' ');
      Label.ChangeChAll('\'', ' ');
      double Size = NIDSizeH.IsKey(NID)? NIDSizeH.GetDat(NID).Val: 2.0;
      double Alpha = 1.0;
      fprintf(F, "\t\t\t<node id='%d' label='%s'>\n", NID, Label.CStr());
      fprintf(F, "\t\t\t\t<viz:color r='%d' g='%d' b='%d' a='%.1f'/>\n", Color.Val1.Val, Color.Val2.Val, Color.Val3.Val, Alpha);
      fprintf(F, "\t\t\t\t<viz:size value='%.3f'/>\n", Size);
      //specify attributes
      if (NIDComVHAtr.IsKey(NID)) {
        fprintf(F, "\t\t\t\t<attvalues>\n");
        for (int c = 0; c < NIDComVHAtr.GetDat(NID).Len(); c++) {
          int CID = NIDComVHAtr.GetDat(NID)[c];
          fprintf(F, "\t\t\t\t\t<attvalue for='%d' value='true'/>\n", CID);
        }
        fprintf(F, "\t\t\t\t</attvalues>\n");
      }

      fprintf(F, "\t\t\t</node>\n");
    }
    fprintf(F, "\t\t</nodes>\n");
    //plot edges
    int EID = 0;
    fprintf(F, "\t\t<edges>\n");
    for (typename PGraph::TObj::TNodeI NI = G->BegNI(); NI < G->EndNI(); NI++) {
      if (NI.GetOutDeg() == 0 && NI.GetInDeg() == 0  ) { continue; }
      for (int e = 0; e < NI.GetOutDeg(); e++) {
        if (! GraphType && NI.GetId() > NI.GetOutNId(e)) { continue; }
        TIntPr Pair = TIntPr(NI.GetId(), NI.GetOutNId(e));
        TIntTr Color = EIDColorH.IsKey(Pair)? EIDColorH.GetDat(Pair) : TIntTr(180, 180, 180);
        double Size = EIDSizeH.IsKey(Pair)? EIDSizeH.GetDat(Pair).Val: 1.0;
        double Alpha = 1.0;
        fprintf(F, "\t\t\t<edge id='%d' source='%d' target='%d'>\n", EID++, NI.GetId(), NI.GetOutNId(e));
        fprintf(F, "\t\t\t\t<viz:color r='%d' g='%d' b='%d' a='%.1f'/>\n", Color.Val1.Val, Color.Val2.Val, Color.Val3.Val, Alpha);
        fprintf(F, "\t\t\t\t<viz:size value='%.3f'/>\n", Size);
        fprintf(F, "\t\t\t</edge>\n");
      }
    }
    fprintf(F, "\t\t</edges>\n");
    fprintf(F, "\t</graph>\n");
    fprintf(F, "</gexf>\n");
    fclose(F);
  }
  static void SaveBipartiteGephi(const TStr& OutFNm, const TIntV& NIDV, const TVec<TIntV>& CmtyVV, const double MaxSz, const double MinSz, const TIntStrH& NIDNameH, const THash<TInt, TIntTr >& NIDColorH, const THash<TInt, TIntTr >& CIDColorH);
    template <class PGraph>
  static PGraph LoadEdgeListStr(const TStr& InFNm, TIntStrH& NIDNameH, const int& SrcColId = 0, const int& DstColId = 1, const TSsFmt SsFmt = ssfTabSep) {
    TSsParser Ss(InFNm, SsFmt);
    PGraph Graph = PGraph::TObj::New();
    TStrHash<TInt> StrSet(Mega(1), true);
    while (Ss.Next()) {
      const int SrcNId = StrSet.AddKey(Ss[SrcColId]);
      const int DstNId = StrSet.AddKey(Ss[DstColId]);
      if (! Graph->IsNode(SrcNId)) { Graph->AddNode(SrcNId); }
      if (! Graph->IsNode(DstNId)) { Graph->AddNode(DstNId); }
      Graph->AddEdge(SrcNId, DstNId);
    }
    NIDNameH.Gen(StrSet.Len());
    for (int s = 0; s < StrSet.Len(); s++) { NIDNameH.AddDat(s, StrSet.GetKey(s)); }
    IAssert(NIDNameH.Len() == Graph->GetNodes());

    Graph->Defrag();
    return Graph;
  }

  
  template<class PGraph>
  static void GVizComGraph(const PGraph& Graph,const TVec<TIntV >& CmtyVV, const TStr& OutFNm, const TStr& Desc = TStr()){
    TStrV Colors = TStrV::GetV("red","blue","green","pink","cyan");
    TStrV Shapes = TStrV::GetV("ellipse","triangle","square","pentagon","hexagon");
    THash<TInt,TIntV> NIDComVH;
    GetNodeMembership(NIDComVH,CmtyVV);

    const TStr Ext = OutFNm.GetFExt();
    const TStr GraphFNm = OutFNm.GetSubStr(0, OutFNm.Len() - Ext.Len()) + "dot";
    const bool GraphType = HasGraphFlag(typename PGraph::TObj, gfDirected);
    FILE *F = fopen(GraphFNm.CStr(), "wt");
    if (! Desc.Empty()) fprintf(F, "/*****\n%s\n*****/\n\n", Desc.CStr());
    if (GraphType) { fprintf(F, "digraph G {\n"); } else { fprintf(F, "graph G {\n"); }
    fprintf(F, "  graph [splines=false overlap=false]\n"); //size=\"12,10\" ratio=fill
    fprintf(F, "  node  [width=0.3, height=0.3]\n");
    //nodes
    for (typename PGraph::TObj::TNodeI NI = Graph->BegNI(); NI < Graph->EndNI(); NI++) {
      int NID = NI.GetId();
      TIntV& CIDV = NIDComVH.GetDat(NID);
      IAssert(CIDV.Len()>0);
      TStr ShapeNm = Shapes[(CIDV.Len()-1) % Shapes.Len()];
      TStr ColorNm = Colors[CIDV[0] % Colors.Len()];
      TStr NodeComLabel = TStr::Fmt("%d(",NID);
      for(int i=0;i<CIDV.Len();i++) {
        TStr TmpStr = TStr::Fmt("%d",int(CIDV[i]));
        NodeComLabel += TmpStr;
        if(i<CIDV.Len()-1){NodeComLabel+=",";}
      }
      NodeComLabel += ")";
      fprintf(F, "  %d [style=filled, shape=\"%s\" fillcolor=\"%s\" label=\"%s\"];\n", NI.GetId(), ShapeNm.CStr(),ColorNm.CStr(), NodeComLabel.CStr()); 
    }
  
    // edges
    for (typename PGraph::TObj::TNodeI NI = Graph->BegNI(); NI < Graph->EndNI(); NI++) {
      if (NI.GetOutDeg()==0 && NI.GetInDeg()==0  ) { 
        fprintf(F, "%d;\n", NI.GetId()); }
      else {
        for (int e = 0; e < NI.GetOutDeg(); e++) {
          if (! GraphType && NI.GetId() > NI.GetOutNId(e)) { continue; }
          fprintf(F, "  %d %s %d;\n", NI.GetId(), GraphType?"->":"--", NI.GetOutNId(e)); 
        }
      }
    }
    if (! Desc.Empty()) {
      fprintf(F, "  label = \"\\n%s\\n\";", Desc.CStr());
      fprintf(F, "  fontsize=24;\n");
    }
    fprintf(F, "}\n");
    fclose(F);
    TSnap::TSnapDetail::GVizDoLayout(GraphFNm, OutFNm, gvlNeato);
  }
};


#endif
